"""AudioFile support for m4a files."""

#
#  Copyright (c) 2000-2025 Stephen Jibson
#
#  This file is part of audiolibrarian.
#
#  Audiolibrarian is free software: you can redistribute it and/or modify it under the terms of the
#  GNU General Public License as published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  Audiolibrarian is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
#  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
#  the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with audiolibrarian.
#  If not, see <https://www.gnu.org/licenses/>.
#
from logging import getLogger
from typing import Any

import mutagen.mp4

from audiolibrarian import audiofile, records

log = getLogger(__name__)
ITUNES = "----:com.apple.iTunes"


class M4aFile(audiofile.AudioFile, extensions={".m4a"}):
    """AudioFile for M4A files."""

    def read_tags(self) -> records.OneTrack:
        """Read the tags and return a OneTrack object."""

        def get_str(key: str) -> str | None:
            # Return first element for the given key, utf8-decoded.
            if mut.get(key) is None:
                return None
            return str(mut.get(key)[0].decode("utf8"))

        def get_strl(key: str) -> records.ListF | None:
            # Return all elements for a given key, utf8-decoded.
            if mut.get(key) is None:
                return None
            return records.ListF([x.decode("utf8") for x in mut.get(key)])

        def listf(key: str) -> records.ListF | None:
            # Return a ListF object for a given key.
            if mut.get(key) is None:
                return None
            return records.ListF(mut.get(key))

        mut = self._mut_file

        front_cover = None
        if mut.get("covr"):
            cover = mut["covr"][0]
            # noinspection PyUnresolvedReferences
            mime = (
                "image/png" if cover.imageformat == mutagen.mp4.AtomDataType.PNG else "image/jpg"
            )
            front_cover = records.FrontCover(data=bytes(cover), mime=mime)
        medium_count = int(mut["disk"][0][1]) if mut.get("disk") else None
        medium_number = int(mut["disk"][0][0]) if mut.get("disk") else None
        track_count = int(mut["trkn"][0][1]) if mut.get("trkn") else None
        track_number = int(mut["trkn"][0][0]) if mut.get("trkn") else None
        release = (
            records.Release(
                album=mut.get("\xa9alb", [None])[0],
                album_artists=listf("aART"),
                album_artists_sort=listf("soaa"),
                asins=get_strl(f"{ITUNES}:ASIN"),
                barcodes=get_strl(f"{ITUNES}:BARCODE"),
                catalog_numbers=get_strl(f"{ITUNES}:CATALOGNUMBER"),
                date=mut.get("\xa9day", [None])[0],
                front_cover=front_cover,
                genres=listf("\xa9gen"),
                labels=get_strl(f"{ITUNES}:LABEL"),
                media={
                    medium_number: records.Medium(
                        formats=get_strl(f"{ITUNES}:MEDIA"),
                        titles=get_strl(f"{ITUNES}:DISCSUBTITLE"),
                        track_count=track_count,
                        tracks={
                            track_number: records.Track(
                                artist=mut.get("\xa9ART", [None])[0],
                                artists=get_strl(f"{ITUNES}:ARTISTS"),
                                artists_sort=mut.get("soar"),
                                file_info=records.FileInfo(
                                    bitrate=mut.info.bitrate // 1000,
                                    bitrate_mode=records.BitrateMode.CBR,
                                    path=self.filepath,
                                    type=records.FileType.AAC,
                                ),
                                isrcs=get_strl(f"{ITUNES}:ISRC"),
                                musicbrainz_artist_ids=get_strl(f"{ITUNES}:MusicBrainz Artist Id"),
                                musicbrainz_release_track_id=get_str(
                                    f"{ITUNES}:MusicBrainz Release Track Id"
                                ),
                                musicbrainz_track_id=get_str(f"{ITUNES}:MusicBrainz Track Id"),
                                title=mut.get("\xa9nam", [None])[0],
                                track_number=int(mut["trkn"][0][0]) if mut.get("trkn") else None,
                            )
                        }
                        if track_number
                        else None,
                    )
                }
                if medium_number
                else None,
                medium_count=medium_count,
                musicbrainz_album_artist_ids=get_strl(f"{ITUNES}:MusicBrainz Album Artist Id"),
                musicbrainz_album_id=get_str(f"{ITUNES}:MusicBrainz Album Id"),
                musicbrainz_release_group_id=get_str(f"{ITUNES}:MusicBrainz Release Group Id"),
                original_date=get_str(f"{ITUNES}:originaldate"),
                original_year=get_str(f"{ITUNES}:originalyear") or None,
                people=(
                    records.People(
                        arrangers=get_strl(f"{ITUNES}:ARRANGER"),
                        composers=get_strl(f"{ITUNES}:COMPOSER"),
                        conductors=get_strl(f"{ITUNES}:CONDUCTOR"),
                        engineers=get_strl(f"{ITUNES}:ENGINEER"),
                        lyricists=get_strl(f"{ITUNES}:LYRICIST"),
                        mixers=get_strl(f"{ITUNES}:MIXER"),
                        producers=get_strl(f"{ITUNES}:PRODUCER"),
                        performers=None,
                        writers=get_strl(f"{ITUNES}:WRITER"),
                    )
                    or None
                ),
                release_countries=get_strl(f"{ITUNES}:MusicBrainz Album Release Country"),
                release_statuses=get_strl(f"{ITUNES}:MusicBrainz Album Status"),
                release_types=get_strl(f"{ITUNES}:MusicBrainz Album Type"),
                script=get_str(f"{ITUNES}:SCRIPT"),
            )
            or None
        )
        if release:
            release.source = records.Source.TAGS
        return records.OneTrack(
            release=release, medium_number=medium_number, track_number=track_number
        )

    def write_tags(self) -> None:
        """Write the tags."""

        def ff(text: int | str | None) -> bytes | None:
            if text is None:
                return None
            return mutagen.mp4.MP4FreeForm(bytes(str(text), "utf8"))  # type: ignore[no-untyped-call]

        def ffl(list_: list[str] | None | Any) -> records.ListF | None:  # noqa: ANN401
            if not list_:
                return None
            return records.ListF([ff(x) for x in list_])

        # Note: We don't write "performers" to m4a files.
        release, medium_number, medium, track_number, track = self._get_tag_sources()
        front_cover = None
        if (cover := release.front_cover) is not None:
            # noinspection PyUnresolvedReferences
            image_format = (
                mutagen.mp4.AtomDataType.PNG
                if cover.mime == "image/png"
                else mutagen.mp4.AtomDataType.JPEG
            )
            front_cover = [
                mutagen.mp4.MP4Cover(cover.data, imageformat=image_format)  # type: ignore[no-untyped-call]
            ]
        tags_ = {
            f"{ITUNES}:ARRANGER": ffl(release.people and release.people.arrangers),
            f"{ITUNES}:ARTISTS": ffl(track.artists),
            f"{ITUNES}:ASIN": ffl(release.asins),
            f"{ITUNES}:BARCODE": ffl(release.barcodes),
            f"{ITUNES}:CATALOGNUMBER": ffl(release.catalog_numbers),
            f"{ITUNES}:COMPOSER": ffl(release.people and release.people.composers),
            f"{ITUNES}:CONDUCTOR": ffl(release.people and release.people.conductors),
            f"{ITUNES}:DISCSUBTITLE": ffl(medium.titles),
            f"{ITUNES}:ENGINEER": ffl(release.people and release.people.engineers),
            f"{ITUNES}:ISRC": ffl(track.isrcs),
            f"{ITUNES}:LABEL": ffl(release.labels),
            f"{ITUNES}:LYRICIST": ffl(release.people and release.people.lyricists),
            f"{ITUNES}:MEDIA": ffl(medium.formats),
            f"{ITUNES}:MIXER": ffl(release.people and release.people.mixers),
            f"{ITUNES}:MusicBrainz Album Artist Id": ffl(release.musicbrainz_album_artist_ids),
            f"{ITUNES}:MusicBrainz Album Id": [ff(release.musicbrainz_album_id)],
            f"{ITUNES}:MusicBrainz Album Release Country": ffl(release.release_countries),
            f"{ITUNES}:MusicBrainz Album Status": ffl(release.release_statuses),
            f"{ITUNES}:MusicBrainz Album Type": ffl(release.release_types),
            f"{ITUNES}:MusicBrainz Artist Id": ffl(track.musicbrainz_artist_ids),
            f"{ITUNES}:MusicBrainz Release Group Id": [ff(release.musicbrainz_release_group_id)],
            f"{ITUNES}:MusicBrainz Release Track Id": [ff(track.musicbrainz_release_track_id)],
            f"{ITUNES}:MusicBrainz Track Id": [ff(track.musicbrainz_track_id)],
            f"{ITUNES}:originaldate": [ff(release.original_date)],
            f"{ITUNES}:originalyear": [ff(release.original_year)],
            f"{ITUNES}:PRODUCER": ffl(release.people and release.people.producers),
            f"{ITUNES}:SCRIPT": [ff(release.script)],
            f"{ITUNES}:WRITER": ffl(release.people and release.people.writers),
            "\xa9alb": [release.album],
            "\xa9ART": [track.artist],
            "\xa9day": [release.date],
            "\xa9gen": release.genres,
            "\xa9nam": [track.title],
            "aART": release.album_artists,
            "covr": front_cover,
            "disk": [(medium_number, release.medium_count)] if medium_number else None,
            "soaa": release.album_artists_sort,
            "soar": track.artists_sort,
            "trkn": [(track_number, medium.track_count)] if track_number else None,
        }
        tags_ = audiofile.Tags(tags_)

        for key, value in tags_.items():
            try:
                self._mut_file[key] = value
            except Exception:  # pragma: no cover
                log.critical("ERROR: %s %s", key, value)
                raise

        self._mut_file.save()



"""AudioFile support for flac files."""

#
#  Copyright (c) 2000-2025 Stephen Jibson
#
#  This file is part of audiolibrarian.
#
#  Audiolibrarian is free software: you can redistribute it and/or modify it under the terms of the
#  GNU General Public License as published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  Audiolibrarian is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
#  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
#  the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with audiolibrarian.
#  If not, see <https://www.gnu.org/licenses/>.
#
import re
from typing import Any

import mutagen.flac

from audiolibrarian import audiofile, records


class FlacFile(audiofile.AudioFile, extensions={".flac"}):
    """AudioFile for Flac files."""

    def read_tags(self) -> records.OneTrack:
        """Read the tags and return a OneTrack object."""

        def listf(lst: list[Any] | None) -> records.ListF | None:
            if lst is None:
                return None
            return records.ListF(lst)

        mut = self._mut_file
        front_cover = None
        if self._mut_file.pictures:
            cover = self._mut_file.pictures[0]
            front_cover = records.FrontCover(
                data=cover.data, desc=cover.desc or "", mime=cover.mime
            )
        medium_count = int(mut["disctotal"][0]) if mut.get("disctotal") else None
        medium_number = int(mut["discnumber"][0]) if mut.get("discnumber") else None
        track_count = int(mut["tracktotal"][0]) if mut.get("tracktotal") else None
        track_number = int(mut["tracknumber"][0]) if mut.get("tracknumber") else None
        release = (
            records.Release(
                album=mut.get("album", [None])[0],
                album_artists=listf(mut.get("albumartist")),
                album_artists_sort=listf(mut.get("albumartistsort")),
                asins=mut.get("asin"),
                barcodes=mut.get("barcode"),
                catalog_numbers=mut.get("catalognumber"),
                date=mut.get("date", [None])[0],
                front_cover=front_cover,
                genres=listf(mut.get("genre")),
                labels=mut.get("label"),
                media={
                    medium_number: records.Medium(
                        formats=listf(mut.get("media")),
                        titles=mut.get("discsubtitle"),
                        track_count=track_count,
                        tracks={
                            track_number: records.Track(
                                artist=mut.get("artist", [None])[0],
                                artists=listf(mut.get("artists")),
                                artists_sort=mut.get("artistsort"),
                                file_info=records.FileInfo(
                                    bitrate=mut.info.bitrate // 1000,
                                    bitrate_mode=records.BitrateMode.CBR,
                                    path=self.filepath,
                                    type=records.FileType.FLAC,
                                ),
                                isrcs=mut.get("isrc"),
                                musicbrainz_artist_ids=listf(mut.get("musicbrainz_artistid")),
                                musicbrainz_release_track_id=mut.get(
                                    "musicbrainz_releasetrackid", [None]
                                )[0],
                                musicbrainz_track_id=mut.get("musicbrainz_trackid", [None])[0],
                                title=mut.get("title", [None])[0],
                                track_number=track_number,
                            )
                        }
                        if track_number
                        else None,
                    )
                }
                if medium_number
                else None,
                medium_count=medium_count,
                musicbrainz_album_artist_ids=listf(mut.get("musicbrainz_albumartistid")),
                musicbrainz_album_id=mut.get("musicbrainz_albumid", [None])[0],
                musicbrainz_release_group_id=mut.get("musicbrainz_releasegroupid", [None])[0],
                original_date=mut.get("originaldate", [None])[0],
                original_year=mut["originalyear"][0] if mut.get("originalyear") else None,
                people=(
                    records.People(
                        arrangers=mut.get("arranger"),
                        composers=mut.get("composer"),
                        conductors=mut.get("conductor"),
                        engineers=mut.get("engineer"),
                        lyricists=mut.get("lyricist"),
                        mixers=mut.get("mixer"),
                        performers=mut.get("performer")
                        and self._parse_performer_tag(mut["performer"]),
                        producers=mut.get("producer"),
                        writers=mut.get("writer"),
                    )
                    or None
                ),
                release_countries=mut.get("releasecountry"),
                release_statuses=mut.get("releasestatus"),
                release_types=mut.get("releasetype"),
                script=mut.get("script", [None])[0],
            )
            or None
        )
        if release:
            release.source = records.Source.TAGS
        return records.OneTrack(
            release=release, medium_number=medium_number, track_number=track_number
        )

    def write_tags(self) -> None:
        """Write the tags."""
        release, medium_number, medium, track_number, track = self._get_tag_sources()
        tags_ = {
            "album": [release.album],
            "albumartist": release.album_artists,
            "albumartistsort": release.album_artists_sort,
            "arranger": release.people and release.people.arrangers,
            "artist": [track.artist],
            "artists": track.artists,
            "artistsort": track.artists_sort,
            "asin": release.asins,
            "barcode": release.barcodes,
            "catalognumber": release.catalog_numbers,
            "composer": release.people and release.people.composers,
            "conductor": release.people and release.people.conductors,
            "date": [release.date],
            "discnumber": [str(medium_number)],
            "discsubtitle": medium.titles,
            "disctotal": [str(release.medium_count)],
            "engineer": release.people and release.people.engineers,
            "genre": release.genres,
            "isrc": track.isrcs,
            "label": release.labels,
            "lyricist": release.people and release.people.lyricists,
            "media": medium.formats,
            "mixer": release.people and release.people.mixers,
            "musicbrainz_albumartistid": release.musicbrainz_album_artist_ids,
            "musicbrainz_albumid": [release.musicbrainz_album_id],
            "musicbrainz_artistid": track.musicbrainz_artist_ids,
            "musicbrainz_releasegroupid": [release.musicbrainz_release_group_id],
            "musicbrainz_releasetrackid": [track.musicbrainz_release_track_id],
            "musicbrainz_trackid": [track.musicbrainz_track_id],
            "originaldate": [release.original_date],
            "originalyear": [str(release.original_year)],
            "performer": self._make_performer_tag(release.people and release.people.performers),
            "producer": release.people and release.people.producers,
            "releasecountry": release.release_countries,
            "releasestatus": release.release_statuses,
            "releasetype": release.release_types,
            "script": [release.script],
            "title": [track.title],
            "totaldiscs": [str(release.medium_count)],
            "totaltracks": [str(medium.track_count)],
            "tracknumber": [str(track_number)],
            "tracktotal": [str(medium.track_count)],
            "writer": release.people and release.people.writers,
        }
        tags_ = audiofile.Tags(tags_)
        self._mut_file.delete()  # Clear old tags.
        self._mut_file.clear_pictures()
        self._mut_file.update(tags_)

        if release.front_cover is not None:
            cover = mutagen.flac.Picture()  # type: ignore[no-untyped-call]
            cover.type = 3
            cover.mime = release.front_cover.mime
            cover.desc = release.front_cover.desc or ""
            cover.data = release.front_cover.data
            self._mut_file.add_picture(cover)

        self._mut_file.save()

    @staticmethod
    def _make_performer_tag(performers: list[records.Performer] | None | Any) -> list[str] | None:  # noqa: ANN401
        # Return a list of performer tag strings "name (instrument)".
        if performers is None:
            return None
        return [f"{p.name} ({p.instrument})" for p in performers]

    @staticmethod
    def _parse_performer_tag(performers_tag: list[str]) -> list[records.Performer]:
        # Parse a list of performer tags and return a list of Performer objects.
        performer_re = re.compile(r"(?P<name>.*)\((?P<instrument>.*)\)")
        performers = []
        for tag in performers_tag:
            if match := performer_re.match(tag):
                name = match.groupdict()["name"].strip()
                instrument = match.groupdict()["instrument"].strip()
                performers.append(records.Performer(name=name, instrument=instrument))
        return performers



"""AudioFile support for mp3 files."""

#
#  Copyright (c) 2000-2025 Stephen Jibson
#
#  This file is part of audiolibrarian.
#
#  Audiolibrarian is free software: you can redistribute it and/or modify it under the terms of the
#  GNU General Public License as published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  Audiolibrarian is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
#  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
#  the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with audiolibrarian.
#  If not, see <https://www.gnu.org/licenses/>.
#
from typing import Any, no_type_check

import mutagen
import mutagen.id3

from audiolibrarian import audiofile, records

APIC = mutagen.id3.APIC
TXXX = mutagen.id3.TXXX
UFID = mutagen.id3.UFID
MB_UFID = "http://musicbrainz.org"


class Mp3File(audiofile.AudioFile, extensions={".mp3"}):
    """AudioFile for MP3 files."""

    @no_type_check  # The mutagen library doesn't provide type hints.
    def read_tags(self) -> records.OneTrack:
        """Read the tags and return a OneTrack object."""

        def get_l(key: str) -> records.ListF | None:
            if (value := mut.get(key)) is None:
                return None
            if "/" in str(value):
                return records.ListF(str(value).split("/"))
            return records.ListF(value)

        mut = self._mut_file
        front_cover = None
        if apic := mut.get("APIC:front cover", mut.get("APIC:front", mut.get("APIC:"))):
            front_cover = records.FrontCover(data=apic.data, desc=apic.desc, mime=apic.mime)
        tipl = mut["TIPL"].people if mut.get("TIPL") else []
        roles = ("arranger", "composer", "conductor", "engineer", "mix", "producer", "writer")
        medium_count = int(mut["TPOS"][0].split("/")[1]) if mut.get("TPOS") else None
        medium_number = int(mut["TPOS"][0].split("/")[0]) if mut.get("TPOS") else None
        track_count = int(mut["TRCK"][0].split("/")[1]) if mut.get("TRCK") else None
        track_number = int(mut["TRCK"][0].split("/")[0]) if mut.get("TRCK") else None
        bitrate = mut.info.bitrate // 1000
        bitrate_mode = records.BitrateMode.__members__[
            str(mut.info.bitrate_mode).rsplit(".", maxsplit=1)[-1]
        ]
        # Shortcut for common CBRs.
        if bitrate_mode == records.BitrateMode.UNKNOWN and bitrate in (128, 160, 192, 320):
            bitrate_mode = records.BitrateMode.CBR
        release = (
            records.Release(
                album=mut.get("TALB", [None])[0],
                album_artists=get_l("TPE2"),
                album_artists_sort=get_l("TSO2"),
                asins=get_l("TXXX:ASIN"),
                barcodes=get_l("TXXX:BARCODE"),
                catalog_numbers=get_l("TXXX:CATALOGNUMBER"),
                date=(mut.get("TDRC", [mutagen.id3.ID3TimeStamp("")])[0]).text or None,
                front_cover=front_cover,
                genres=get_l("TCON"),
                labels=get_l("TPUB"),
                media={
                    medium_number: records.Medium(
                        formats=get_l("TMED"),
                        titles=get_l("TSST"),
                        track_count=track_count,
                        tracks={
                            track_number: records.Track(
                                artist=mut.get("TPE1", [None])[0],
                                artists=get_l("TXXX:ARTISTS"),
                                artists_sort=get_l("TSOP"),
                                file_info=records.FileInfo(
                                    bitrate=bitrate,
                                    bitrate_mode=bitrate_mode,
                                    path=self.filepath,
                                    type=records.FileType.MP3,
                                ),
                                isrcs=get_l("TSRC"),
                                musicbrainz_artist_ids=get_l("TXXX:MusicBrainz Artist Id"),
                                musicbrainz_release_track_id=mut.get(
                                    "TXXX:MusicBrainz Release Track Id", [None]
                                )[0],
                                musicbrainz_track_id=mut.get(
                                    f"UFID:{MB_UFID}", UFID()
                                ).data.decode("utf8")
                                or None,
                                title=mut.get("TIT2", [None])[0],
                                track_number=track_number,
                            )
                        }
                        if track_number
                        else None,
                    )
                }
                if medium_number
                else None,
                medium_count=medium_count,
                musicbrainz_album_artist_ids=get_l("TXXX:MusicBrainz Album Artist Id"),
                musicbrainz_album_id=mut.get("TXXX:MusicBrainz Album Id", [None])[0],
                musicbrainz_release_group_id=mut.get("TXXX:MusicBrainz Release Group Id", [None])[
                    0
                ],
                original_year=str((mut["TDOR"][0]).year) if mut.get("TDOR") else None,
                people=(
                    records.People(
                        arrangers=[name for role, name in tipl if role == "arranger"] or None,
                        composers=[name for role, name in tipl if role == "composer"] or None,
                        conductors=[name for role, name in tipl if role == "conductor"] or None,
                        engineers=[name for role, name in tipl if role == "engineer"] or None,
                        lyricists=get_l("TEXT"),
                        mixers=[name for role, name in tipl if role == "mix"] or None,
                        producers=[name for role, name in tipl if role == "producer"] or None,
                        performers=[
                            records.Performer(name=n, instrument=i)
                            for i, n in tipl
                            if i not in roles
                        ]
                        or None,
                        writers=[name for role, name in tipl if role == "writer"] or None,
                    )
                    or None
                ),
                release_countries=get_l("TXXX:MusicBrainz Album Release Country"),
                release_statuses=get_l("TXXX:MusicBrainz Album Status"),
                release_types=get_l("TXXX:MusicBrainz Album Type"),
                script=mut.get("TXXX:SCRIPT", [None])[0],
            )
            or None
        )
        if release:
            release.source = records.Source.TAGS
        return records.OneTrack(
            release=release, medium_number=medium_number, track_number=track_number
        )

    @no_type_check  # The mutagen library doesn't provide type hints.
    def write_tags(self) -> None:  # noqa: C901, PLR0912, PLR0915
        """Write the tags."""

        def slash(text: list[str] | records.ListF) -> str:
            return "/".join(text)

        release, medium_number, medium, track_number, track = self._get_tag_sources()
        tipl_people = (
            [["arranger", x] for x in (release.people and release.people.arrangers) or []]
            + [["composer", x] for x in (release.people and release.people.composers) or []]
            + [["conductor", x] for x in (release.people and release.people.conductors) or []]
            + [["engineer", x] for x in (release.people and release.people.engineers) or []]
            + [["mix", x] for x in (release.people and release.people.mixers) or []]
            + [["producer", x] for x in (release.people and release.people.producers) or []]
            + [
                [p.instrument, p.name]
                for p in (release.people and release.people.performers) or []
            ]
            + [["writer", x] for x in (release.people and release.people.writers) or []]
        )
        tags: list[mutagen.id3.Frame] = []
        # noinspection PyUnusedLocal
        tag: Any = None
        if tag := release.album:
            tags.append(mutagen.id3.TALB(encoding=1, text=tag))
        if tag := release.people and release.people.composers:
            tags.append(mutagen.id3.TCOM(encoding=1, text=slash(tag)))
        if tag := release.genres:
            tags.append(mutagen.id3.TCON(encoding=3, text=slash(tag)))
        if tag := release.original_year:
            tags.append(mutagen.id3.TDOR(encoding=0, text=str(tag)))
        if tag := release.date:
            tags.append(mutagen.id3.TDRC(encoding=0, text=tag))
        if tag := release.people and release.people.lyricists:
            tags.append(mutagen.id3.TEXT(encoding=1, text=slash(tag)))
        if people := tipl_people:
            tags.append(mutagen.id3.TIPL(encoding=1, people=people))
        if tag := track.title:
            tags.append(mutagen.id3.TIT2(encoding=1, text=tag))
        if tag := medium.formats:
            tags.append(mutagen.id3.TMED(encoding=1, text=slash(tag)))
        if tag := track.artist:
            tags.append(mutagen.id3.TPE1(encoding=1, text=tag))
        if tag := release.album_artists:
            tags.append(mutagen.id3.TPE2(encoding=1, text=slash(tag)))
        if tag := release.people and release.people.conductors:
            tags.append(mutagen.id3.TPE3(encoding=1, text=slash(tag)))
        if (num := medium_number) and (tag := release.medium_count):
            tags.append(mutagen.id3.TPOS(encoding=0, text=f"{num}/{tag}"))
        if tag := release.labels:
            tags.append(mutagen.id3.TPUB(encoding=1, text=slash(tag)))
        if (num := track_number) and (tag := medium.track_count):
            tags.append(mutagen.id3.TRCK(encoding=0, text=f"{num}/{tag}"))
        if tag := release.album_artists_sort:
            tags.append(mutagen.id3.TSO2(encoding=1, text=slash(tag)))
        if tag := track.artists_sort:
            tags.append(mutagen.id3.TSOP(encoding=1, text=slash(tag)))
        if tag := track.isrcs:
            tags.append(mutagen.id3.TSRC(encoding=1, text=slash(tag)))
        if tag := medium.titles:
            tags.append(mutagen.id3.TSST(encoding=1, text=slash(tag)))
        if tag := track.artists:
            tags.append(TXXX(encoding=1, desc="ARTISTS", text=slash(tag)))
        if tag := release.asins:
            tags.append(TXXX(encoding=1, desc="ASIN", text=slash(tag)))
        if tag := release.barcodes:
            tags.append(TXXX(encoding=1, desc="BARCODE", text=slash(tag)))
        if tag := release.catalog_numbers:
            tags.append(TXXX(encoding=1, desc="CATALOGNUMBER", text=slash(tag)))
        if tag := release.musicbrainz_album_artist_ids:
            tags.append(TXXX(encoding=1, desc="MusicBrainz Album Artist Id", text=slash(tag)))
        if tag := release.musicbrainz_album_id:
            tags.append(TXXX(encoding=1, desc="MusicBrainz Album Id", text=tag))
        if tag := release.release_countries:
            tags.append(
                TXXX(encoding=1, desc="MusicBrainz Album Release Country", text=slash(tag))
            )
        if tag := release.release_statuses:
            tags.append(TXXX(encoding=1, desc="MusicBrainz Album Status", text=slash(tag)))
        if tag := release.release_types:
            tags.append(TXXX(encoding=1, desc="MusicBrainz Album Type", text=slash(tag)))
        if tag := track.musicbrainz_artist_ids:
            tags.append(TXXX(encoding=1, desc="MusicBrainz Artist Id", text=slash(tag)))
        if tag := release.musicbrainz_release_group_id:
            tags.append(TXXX(encoding=1, desc="MusicBrainz Release Group Id", text=tag))
        if tag := track.musicbrainz_release_track_id:
            tags.append(TXXX(encoding=1, desc="MusicBrainz Release Track Id", text=tag))
        if tag := release.script:
            tags.append(TXXX(encoding=1, desc="SCRIPT", text=tag))
        if tag := release.original_year:
            tags.append(TXXX(encoding=1, desc="originalyear", text=str(tag)))
        if id_ := track.musicbrainz_track_id:
            tags.append(UFID(owner=MB_UFID, data=bytes(id_, "utf8")))
        if (cover := release.front_cover) is not None:
            tags.append(
                APIC(encoding=0, mime=cover.mime, type=3, desc=cover.desc, data=cover.data)
            )

        try:
            id3 = mutagen.id3.ID3(self.filepath)
        except mutagen.id3.ID3NoHeaderError:
            self._mut_file.add_tags()
            self._mut_file.save()
            id3 = mutagen.id3.ID3(self.filepath)

        id3.delete()
        for tag in tags:
            id3.add(tag)
        id3.save()
        self._mut_file = mutagen.File(self.filepath.absolute())



"""Screen output utilities."""

#
#  Copyright (c) 2000-2025 Stephen Jibson
#
#  This file is part of audiolibrarian.
#
#  Audiolibrarian is free software: you can redistribute it and/or modify it under the terms of the
#  GNU General Public License as published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  Audiolibrarian is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
#  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
#  the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with audiolibrarian.
#  If not, see <https://www.gnu.org/licenses/>.
#
import sys
from types import TracebackType


class Dots:
    """Context Manager that outputs a message, and dots...

    Example:
        with Dots("Please wait...") as d:
            for _ in range(10):
                time.sleep(1)  # or better, actually do some work here instead
                d.dot()
    """

    def __init__(self, message: str) -> None:
        """Initialize a Dots object."""
        self._out(message)

    def __enter__(self) -> "Dots":
        """Enter the context manager."""
        return self

    def __exit__(
        self,
        _: type[BaseException] | None,
        __: BaseException | None,
        ___: TracebackType | None,
    ) -> None:
        """Exit the context manager."""
        self._out("\n")

    def dot(self) -> None:
        """Output a dot."""
        self._out(".")

    @staticmethod
    def _out(message: str) -> None:
        sys.stdout.write(message)
        sys.stdout.flush()



"""AudioLibrarian base class.

Useful stuff: https://help.mp3tag.de/main_tags.html
"""

#
#  Copyright (c) 2000-2025 Stephen Jibson
#
#  This file is part of audiolibrarian.
#
#  Audiolibrarian is free software: you can redistribute it and/or modify it under the terms of the
#  GNU General Public License as published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  Audiolibrarian is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
#  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
#  the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with audiolibrarian.
#  If not, see <https://www.gnu.org/licenses/>.
#
import argparse
import logging
import pathlib
import shutil
import subprocess
import sys
import warnings
from collections.abc import Iterable
from typing import Any, Final

import colors
import ffmpeg_normalize
import filelock
import yaml

from audiolibrarian import audiofile, audiosource, musicbrainz, records, sh, text
from audiolibrarian.settings import SETTINGS

log = logging.getLogger(__name__)


class Base:
    """AudioLibrarian base class.

    This class should be sub-classed for various sub_commands.
    """

    command: str | None = None
    _manifest_file: Final[str] = "Manifest.yaml"

    def __init__(self, args: argparse.Namespace) -> None:
        """Initialize the base."""
        # Pull in stuff from args.
        search_keys = ("album", "artist", "mb_artist_id", "mb_release_id")
        self._provided_search_data = {k: v for k, v in vars(args).items() if k in search_keys}
        if vars(args).get("disc"):
            self._disc_number, self._disc_count = [int(x) for x in args.disc.split("/")]
        else:
            self._disc_number, self._disc_count = 1, 1

        # Directories.
        self._library_dir = SETTINGS.library_dir
        self._work_dir = SETTINGS.work_dir
        self._flac_dir = self._work_dir / "flac"
        self._m4a_dir = self._work_dir / "m4a"
        self._mp3_dir = self._work_dir / "mp3"
        self._source_dir = self._work_dir / "source"
        self._wav_dir = self._work_dir / "wav"

        self._lock = filelock.FileLock(str(self._work_dir) + ".lock")

        self._normalizer = self._which_normalizer()

        # Initialize stuff that will be defined later.
        self._audio_source: audiosource.AudioSource | None = None
        self._release: records.Release | None = None
        self._medium: records.Medium | None = None
        self._source_is_cd: bool | None = None
        self._source_example: records.OneTrack | None = None

    @property
    def _flac_filenames(self) -> list[pathlib.Path]:
        """Return the current list of flac files in the work directory."""
        return sorted(self._flac_dir.glob("*.flac"), key=text.alpha_numeric_key)

    @property
    def _m4a_filenames(self) -> list[pathlib.Path]:
        """Return the current list of m4a files in the work directory."""
        return sorted(self._m4a_dir.glob("*.m4a"), key=text.alpha_numeric_key)

    @property
    def _mp3_filenames(self) -> list[pathlib.Path]:
        """Return the current list of mp3 files in the work directory."""
        return sorted(self._mp3_dir.glob("*.mp3"), key=text.alpha_numeric_key)

    @property
    def _multi_disc(self) -> bool:
        """Return True if this is part of a multi-disc set."""
        return (self._disc_number, self._disc_count) != (1, 1)

    @property
    def _source_filenames(self) -> list[pathlib.Path]:
        """Return the current list of source files in the work directory."""
        return sorted(self._source_dir.glob("*.flac"), key=text.alpha_numeric_key)

    @property
    def _wav_filenames(self) -> list[pathlib.Path]:
        """Return the current list of wav files in the work directory."""
        return sorted(self._wav_dir.glob("*.wav"), key=text.alpha_numeric_key)

    def _convert(self, *, make_source: bool = True) -> None:
        """Perform all the steps of ripping, normalizing, converting and moving the files."""
        if self._audio_source is None:
            warnings.warn(
                "Cannot convert; no audio_source is defined.", RuntimeWarning, stacklevel=2
            )
            return
        self._audio_source.prepare_source()
        with self._lock:
            self._make_clean_workdirs()
            self._audio_source.copy_wavs(self._wav_dir)
            self._rename_wav()
            if make_source:
                self._make_source()
            self._normalize()
            self._make_flac()
            self._make_m4a()
            self._make_mp3()
            self._move_files(move_source=make_source)

    def _find_manifests(self, directories: list[str | pathlib.Path]) -> list[pathlib.Path]:
        """Return a sorted, unique list of manifest files anywhere in the given directories."""
        manifests = set()
        for directory in directories:
            path = pathlib.Path(directory)
            for manifest in path.rglob(self._manifest_file):
                manifests.add(manifest)
        return sorted(manifests)

    def _get_searcher(self) -> musicbrainz.Searcher:
        """Return a Searcher object populated with data from the audio source and cli args."""
        search_data: dict[str, str] = (
            self._audio_source.get_search_data() if self._audio_source is not None else {}
        )
        searcher = musicbrainz.Searcher(**search_data)  # type: ignore[arg-type]
        searcher.disc_number = str(self._disc_number)
        # Override with user-provided info.
        if value := self._provided_search_data.get("artist"):
            searcher.artist = value
        if value := self._provided_search_data.get("album"):
            searcher.album = value
        if value := self._provided_search_data.get("mb_artist_id"):
            searcher.mb_artist_id = value
        if value := self._provided_search_data.get("mb_release_id"):
            searcher.mb_release_id = value
        log.info("SEARCHER: %s", searcher)
        return searcher

    def _get_tag_info(self) -> None:
        """Gather search information from user-provided options and source files.

        Set `self._release` and `self._medium` based on the search results.
        If the search results don't have a front cover, use the front cover from the source files.
        Print a summary of the results and prompt the user to confirm.
        If the track count does not match the file count, print an error message.
        If the user does not confirm, exit the program.
        """
        print("Gathering search information...")
        searcher = self._get_searcher()
        skip_confirm = bool(searcher.mb_artist_id and searcher.mb_release_id)
        print("Finding MusicBrainz release information...")
        self._release = searcher.find_music_brains_release()
        self._medium = self._release.media[int(self._disc_number)]
        if (
            not self._release.front_cover
            and self._audio_source
            and (cover := self._audio_source.get_front_cover())
        ):
            log.info("Using front-cover image from source file")
            self._release.front_cover = cover
        summary, okay = self._summary()
        print(summary)
        if not okay:
            print(colors.color("\n*** Track count does not match file count ***\n", fg="red"))
            skip_confirm = False
        if not skip_confirm and text.input_("Confirm [N,y]: ").lower() != "y":  # pragma: no cover
            sys.exit(1)

    def _make_clean_workdirs(self) -> None:
        """Erase everything from the workdir and create the empty directory structure."""
        if self._work_dir.is_dir():
            shutil.rmtree(self._work_dir)
        for path in self._flac_dir, self._m4a_dir, self._mp3_dir, self._source_dir, self._wav_dir:
            path.mkdir(parents=True)

    def _make_flac(self, *, source: bool = False) -> None:
        """Convert the wav files into flac files; tag them.

        If source is True, it stores the flac files in the source directory,
        otherwise, it stores them in the flac directory.
        """
        out_dir = self._source_dir if source else self._flac_dir
        commands: list[tuple[str, ...]] = [
            ("flac", "--silent", f"--output-prefix={out_dir}/", str(f))
            for f in self._wav_filenames
        ]
        sh.parallel(f"Making {len(self._wav_filenames)} flac files...", commands)
        filenames = self._source_filenames if source else self._flac_filenames
        sh.touch(filenames)
        self._tag_files(filenames)

    def _make_m4a(self) -> None:
        """Convert the wav files into m4a files; tag them."""
        commands: list[tuple[str, ...]] = []
        for filename in self._wav_filenames:
            dst_file = self._m4a_dir / filename.name.replace(".wav", ".m4a")
            commands.append(
                ("fdkaac", "--silent", "--bitrate-mode=5", "-o", str(dst_file), str(filename))
            )
        sh.parallel(f"Making {len(commands)} m4a files...", commands)
        sh.touch(self._m4a_filenames)
        self._tag_files(self._m4a_filenames)

    def _make_mp3(self) -> None:
        """Convert the wav files into mp3 files; tag them."""
        commands: list[tuple[str, ...]] = []
        for filename in self._wav_filenames:
            dst_file = self._mp3_dir / filename.name.replace(".wav", ".mp3")
            commands.append(("lame", "--silent", "-h", "-b", "192", str(filename), str(dst_file)))
        sh.parallel(f"Making {len(commands)} mp3 files...", commands)
        sh.touch(self._mp3_filenames)
        self._tag_files(self._mp3_filenames)

    def _make_source(self) -> None:
        """Convert the files into flac files; store them in the source dir; read their tags.

        The files are defined by the audio source; they could be wav files from a CD
        or another type of audio file.
        """
        self._make_flac(source=True)
        self._source_example = audiofile.AudioFile.open(self._source_filenames[0]).read_tags()

    def _move_files(self, *, move_source: bool = True) -> None:
        """Move converted/tagged files from the work directory into the library directory."""
        artist_album_dir = self._release.get_artist_album_path()
        flac_dir = self._library_dir / "flac" / artist_album_dir
        m4a_dir = self._library_dir / "m4a" / artist_album_dir
        mp3_dir = self._library_dir / "mp3" / artist_album_dir
        source_dir = self._library_dir / "source" / artist_album_dir
        if self._multi_disc:
            flac_dir /= f"disc{self._disc_number}"
            m4a_dir /= f"disc{self._disc_number}"
            mp3_dir /= f"disc{self._disc_number}"
            source_dir /= f"disc{self._disc_number}"
        for path in [flac_dir, m4a_dir, mp3_dir] + ([source_dir] if move_source else []):
            if path.is_dir():
                shutil.rmtree(path)
            path.mkdir(parents=True)
        for path in self._flac_filenames:
            path.rename(flac_dir / path.name)
        for path in self._m4a_filenames:
            path.rename(m4a_dir / path.name)
        for path in self._mp3_filenames:
            path.rename(mp3_dir / path.name)
        if move_source:
            for path in self._source_filenames:
                path.rename(source_dir / path.name)

    def _normalize(self) -> None:
        """Normalize the wav files using the selected normalizer."""
        if self._normalizer == "none":
            return
        print(f"Normalizing wav files using {self._normalizer}...")

        if self._normalizer == "wavegain":
            command = [
                "wavegain",
                f"--{SETTINGS.normalize.wavegain.preset}",
                f"--gain={SETTINGS.normalize.wavegain.gain}",
                "--apply",
            ]
            command.extend(str(f) for f in self._wav_filenames)
            result = subprocess.run(command, capture_output=True, check=False)  # noqa: S603
            for line in str(result.stderr).split(r"\n"):
                line_trunc = line[:137] + "..." if len(line) > 140 else line  # noqa: PLR2004
                log.info("WAVEGAIN: %s", line_trunc)
            result.check_returncode()
            return

        normalizer = ffmpeg_normalize.FFmpegNormalize(
            extension="wav",
            keep_loudness_range_target=True,
            target_level=SETTINGS.normalize.ffmpeg.target_level,
        )
        for wav_file in self._wav_filenames:
            normalizer.add_media_file(str(wav_file), str(wav_file))
        log.info("NORMALIZER: starting ffmpeg normalization...")
        normalizer.run_normalization()
        log.info("NORMALIZER: ffmpeg normalization completed successfully")

    def _rename_wav(self) -> None:
        """Rename the wav files to a filename-sane representation of the track title."""
        for old_path in self._wav_filenames:
            track_number = text.get_track_number(str(old_path.name))
            title_filename = self._medium.tracks[track_number].get_filename(".wav")
            new_path = old_path.parent / title_filename
            if new_path.resolve() != old_path.resolve():
                log.info("RENAMING: %s --> %s", old_path.name, new_path.name)
                old_path.rename(new_path)

    def _summary(self) -> tuple[str, bool]:
        """Return a summary of the conversion/tagging process and an "ok" flag indicating issues.

        The summary is a nicely formatted table showing the album, artist and track info.
        The "ok" flag indicating issues will be true if:
          - the file count does not match the song count from the MusicBrainz database
        (https://jrgraphix.net/r/Unicode/2500-257F)
        """
        if self._audio_source is None or self._release is None:
            warnings.warn(
                "Cannot provide summary; missing audio_source and/or release.",
                RuntimeWarning,
                stacklevel=1,
            )
            return "", False
        lines = []
        okay = True
        no_match = "(no match)"
        col1 = [f.stem if f else no_match for f in self._audio_source.source_list]
        col2 = [t.get_filename() for _, t in sorted(self._medium.tracks.items())]
        col3 = [f"{str(n).zfill(2)}: {t.title}" for n, t in sorted(self._medium.tracks.items())]
        min_total_w = 74  # Make sure we've got enough width for MB Release URL.
        width = 40
        col1_w = min(width, max([len(x) for x in col1] + [len(no_match)]))
        col2_w = min(width, max([len(x) for x in col2] + [len(no_match)]))
        col3_w = min(width, max([len(x) for x in col3] + [len(no_match)]))
        if col1_w + col2_w + col3_w < min_total_w:
            col3_w = min_total_w - (col1_w + col2_w)
        tab_w = col1_w + col2_w + col3_w + 6
        c1_line = (col1_w + 2) * "\u2550"
        c2_line = (col2_w + 2) * "\u2550"
        c3_line = (col3_w + 2) * "\u2550"
        alb = f"Album:      {self._release.album}"
        art = f"Artist(s):  {', '.join(self._release.album_artists_sort)}"
        med = f"Disc:       {self._disc_number} of {self._disc_count}"
        mbr = f"MB Release: https://musicbrainz.org/release/{self._release.musicbrainz_album_id}"
        lines.append(f"\u2554{c1_line}\u2550{c2_line}\u2550{c3_line}\u2557")
        lines.extend(
            [f"\u2551 {line} {' ' * (tab_w - len(line))}\u2551" for line in (alb, art, mbr, med)]
        )
        fmt = f"\u2551 {{c1: <{col1_w}}} \u2502 {{c2: <{col2_w}}} \u2502 {{c3: <{col3_w}}} \u2551"
        lines.append(f"\u2560{c1_line}\u2564{c2_line}\u2564{c3_line}\u2563")
        lines.append(fmt.format(c1="Source", c2="Destination", c3="Title"))
        lines.append(f"\u2560{c1_line}\u256a{c2_line}\u256a{c3_line}\u2563")
        rows = max(len(col1), len(col2), len(col3))
        for i in range(rows):
            col1_ = (
                ((col1[i][: width - 3] + "...") if len(col1[i]) > width else col1[i])
                if len(col1) > i
                else no_match
            )
            col2_ = (
                ((col2[i][: width - 3] + "...") if len(col2[i]) > width else col2[i])
                if len(col2) > i
                else no_match
            )
            col3_ = (
                ((col3[i][: width - 3] + "...") if len(col3[i]) > width else col3[i])
                if len(col3) > i
                else no_match
            )
            lines.append(fmt.format(c1=col1_, c2=col2_, c3=col3_))
            if no_match in (col1_, col2_, col3_):
                okay = False
        lines.append(f"\u255a{c1_line}\u2567{c2_line}\u2567{c3_line}\u255d")
        return "\n".join(lines), okay

    def _tag_files(self, filenames: list[pathlib.Path]) -> None:
        """Tag the given list of files."""
        for filename in filenames:
            song = audiofile.AudioFile.open(filename)
            song.one_track = records.OneTrack(
                release=self._release,
                medium_number=self._disc_number,
                track_number=int(filename.name.split("__")[0]),
            )
            song.write_tags()

    def _write_manifest(self) -> None:
        """Write out a manifest file with release information."""
        release = self._release  # We use this a lot below.
        file_info = self._source_example.track.file_info
        manifest = {
            "album": release.album,
            "artist": release.album_artists.first,
            "artist_sort_name": release.album_artists_sort.first,
            "media": release.media[self._disc_number].formats.first,
            "genre": release.genres.first,
            "disc_number": self._disc_number,
            "disc_count": self._disc_count,
            "original_year": release.original_year,
            "date": release.date,
            "musicbrainz_info": {
                "albumid": release.musicbrainz_album_id,
                "albumartistid": release.musicbrainz_album_artist_ids.first,
                "releasegroupid": release.musicbrainz_release_group_id,
            },
            "source_info": {
                "type": file_info.type.name,
                "bitrate": file_info.bitrate,
                "bitrate_mode": file_info.bitrate_mode.name,
            },
        }
        if self.command == "manifest":
            manifest_filename = self._manifest_file  # Write to current directory.
            if self._source_is_cd:
                manifest["source_info"] = {
                    "type": "CD",
                    "bitrate": 1411,
                    "bitrate_mode": "CBR",
                }
        else:
            source_dir = self._library_dir / "source"
            artist_album_dir = self._release.get_artist_album_path()
            manifest_filename = str(source_dir / artist_album_dir / self._manifest_file)
        with pathlib.Path(manifest_filename).open("w", encoding="utf-8") as manifest_file:
            yaml.dump(manifest, manifest_file)
        print(f"Wrote {manifest_filename}")

    @staticmethod
    def _which_normalizer() -> str:
        """Determine which normalizer to use based on settings and availability.

        Returns:
            str: The name of the normalizer to use ("wavegain", "ffmpeg" or "none")
        """
        normalizer = SETTINGS.normalize.normalizer
        if normalizer == "none":
            return "none"

        wavegain_found = shutil.which("wavegain")
        if normalizer in ("auto", "wavegain") and wavegain_found:
            return "wavegain"

        ffmpeg_found = shutil.which("ffmpeg")
        if normalizer in ("auto", "ffmpeg") and ffmpeg_found:
            return "ffmpeg"

        if wavegain_found:
            log.warning("ffmpeg not found, using wavegain for normalization")
            return "wavegain"
        if ffmpeg_found:
            log.warning("wavegain not found, using ffmpeg for normalization")
            return "ffmpeg"
        log.warning("wavegain not found, ffmpeg not found, using no normalization")
        return "none"

    @staticmethod
    def _find_audio_files(directories: list[str | pathlib.Path]) -> Iterable[audiofile.AudioFile]:
        """Yield audiofile objects found in the given directories."""
        paths: list[pathlib.Path] = []
        # Grab all the paths first because thing may change as files are renamed.
        for directory in directories:
            path = pathlib.Path(directory)
            for ext in audiofile.AudioFile.extensions():
                paths.extend(path.rglob(f"*{ext}"))
        paths = sorted(set(paths))
        # Using yield rather than returning a list saves us from simultaneously storing
        # potentially thousands of AudioFile objects in memory at the same time.
        for path in paths:
            try:
                yield audiofile.AudioFile.open(path)
            except FileNotFoundError:
                continue

    @staticmethod
    def _read_manifest(manifest_path: pathlib.Path) -> dict[Any, Any]:
        with manifest_path.open(encoding="utf-8") as manifest_file:
            return dict(yaml.safe_load(manifest_file))



"""Command execution helpers."""

#
#  Copyright (c) 2000-2025 Stephen Jibson
#
#  This file is part of audiolibrarian.
#
#  Audiolibrarian is free software: you can redistribute it and/or modify it under the terms of the
#  GNU General Public License as published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  Audiolibrarian is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
#  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
#  the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with audiolibrarian.
#  If not, see <https://www.gnu.org/licenses/>.
#
import pathlib
import subprocess
from collections.abc import Iterable
from multiprocessing import Pool

from audiolibrarian import output


def _run_command(command: tuple[str, ...]) -> None:
    """Run a single command."""
    subprocess.run(command, check=True)  # noqa: S603


def parallel(
    message: str, commands: list[tuple[str, ...]], max_workers: int | None = None
) -> None:
    """Execute commands in parallel using multiprocessing.

    Args:
        message: Progress message to display
        commands: List of commands to execute
        max_workers: Maximum number of parallel processes (None for system default)
    """
    with output.Dots(message) as dots, Pool(max_workers) as pool:
        # Start all processes
        results = [pool.apply_async(_run_command, (command,)) for command in commands]

        # Wait for all processes to complete
        for result in results:
            result.get()  # Will raise any exceptions from the subprocess
            dots.dot()


def touch(paths: Iterable[pathlib.Path]) -> None:
    """Touch all files in a given path."""
    for path in paths:
        path.touch(exist_ok=True)



"""Manage tags."""

#
#  Copyright (c) 2000-2025 Stephen Jibson
#
#  This file is part of audiolibrarian.
#
#  Audiolibrarian is free software: you can redistribute it and/or modify it under the terms of the
#  GNU General Public License as published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  Audiolibrarian is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
#  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
#  the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with audiolibrarian.
#  If not, see <https://www.gnu.org/licenses/>.
#
from typing import Any


# noinspection PyMissingConstructor
class Tags(dict[Any, Any]):
    """A dict-like object that silently drops keys with None in their values.

    A key will be dropped if:
    * its value is None
    * its value is a list containing None
    * its value is a dict with None in its values
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # noqa: ANN401
        """Initialize a Tags object."""
        self.update(*args, **kwargs)

    def __setitem__(self, k: Any, v: Any) -> None:  # noqa: ANN401
        """Set an item only if it should not be dropped."""
        if not (
            v is None
            or (isinstance(v, list) and (None in v or "None" in v))
            or (isinstance(v, dict) and (None in v.values() or "None" in v.values()))
        ):
            super().__setitem__(k, v)

    def update(self, *args: Any, **kwargs: Any) -> None:  # noqa: ANN401
        """See base class."""
        for key, value in dict(*args, **kwargs).items():
            self[key] = value



"""AudioSource."""

#
#  Copyright (c) 2000-2025 Stephen Jibson
#
#  This file is part of audiolibrarian.
#
#  Audiolibrarian is free software: you can redistribute it and/or modify it under the terms of the
#  GNU General Public License as published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  Audiolibrarian is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
#  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
#  the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with audiolibrarian.
#  If not, see <https://www.gnu.org/licenses/>.
#
import abc
import logging
import os
import pathlib
import shutil
import subprocess
import tempfile
from collections.abc import Callable  # noqa: TC003

import discid

from audiolibrarian import audiofile, records, sh, text
from audiolibrarian.settings import SETTINGS

log = logging.getLogger(__name__)


class AudioSource(abc.ABC):
    """An abstract base class for AudioSource classes."""

    def __init__(self) -> None:
        """Initialize an AudioSource."""
        self._temp_dir: pathlib.Path = pathlib.Path(tempfile.mkdtemp())
        self._source_list: list[pathlib.Path | None] | None = None

    def __del__(self) -> None:
        """Remove any temp files."""
        if self._temp_dir.is_dir():
            shutil.rmtree(self._temp_dir)

    @property
    def source_list(self) -> list[pathlib.Path | None]:
        """Return a list with source file paths and blanks.

        The list will be ordered by track number, with None in spaces where no
        filename is present for that track number.
        """
        if not self._source_list:
            source_filenames = self.get_source_filenames()
            length = max(text.get_track_number(str(f.name)) for f in source_filenames)
            result: list[pathlib.Path | None] = [None] * length
            if length:
                for filename in source_filenames:
                    idx = text.get_track_number(str(filename.name)) - 1
                    result[idx] = filename
            self._source_list = result
        return self._source_list

    def copy_wavs(self, dest_dir: pathlib.Path) -> None:
        """Copy wav files to the given destination directory."""
        for filename in self.get_wav_filenames():
            shutil.copy2(filename, dest_dir / filename.name)

    def get_front_cover(self) -> records.FrontCover | None:
        """Return a FrontCover record or None."""
        return None

    @abc.abstractmethod
    def get_search_data(self) -> dict[str, str]:
        """Return a dictionary of search data useful for doing a MusicBrainz search."""

    @abc.abstractmethod
    def get_source_filenames(self) -> list[pathlib.Path]:
        """Return a list of the original source file paths."""

    def get_wav_filenames(self) -> list[pathlib.Path]:
        """Return a list of the prepared wav file paths."""
        return sorted(self._temp_dir.glob("*.wav"), key=text.alpha_numeric_key)

    @abc.abstractmethod
    def prepare_source(self) -> None:
        """Convert the source to wav files."""


class CDAudioSource(AudioSource):
    """AudioSource from a compact disc."""

    def __init__(self) -> None:
        """Initialize a CDAudioSource."""
        super().__init__()
        self._cd = discid.read(SETTINGS.discid_device, features=["mcn"])

    def get_search_data(self) -> dict[str, str]:
        """Return a dictionary of search data useful for doing a MusicBrainz search."""
        return {"disc_id": self._cd.id, "disc_mcn": self._cd.mcn}

    def get_source_filenames(self) -> list[pathlib.Path]:
        """Return a list of the original source file paths.

        Since we're working with a CD, these files may not yet exist if they have not been
        read from the disc.
        """
        return [
            self._temp_dir / f"track{str(n + 1).zfill(2)}.cdda.wav"
            for n in range(self._cd.last_track_num)
        ]

    def prepare_source(self) -> None:
        """Pull audio from the CD to wav files."""
        cwd = pathlib.Path.cwd()
        os.chdir(self._temp_dir)
        try:
            subprocess.run(("cd-paranoia", "-B"), check=True)
        finally:
            os.chdir(cwd)
        subprocess.run(("eject",), check=False)


class FilesAudioSource(AudioSource):
    """AudioSource from local files."""

    def __init__(self, filenames: list[pathlib.Path]) -> None:
        """Initialize a FilesAudioSource."""
        super().__init__()
        self._filenames = filenames
        if len(filenames) == 1 and filenames[0].is_dir():
            # If we're given a directory, figure out what's in there.
            for file_type in ("flac", "wav", "m4a", "mp3"):
                if fns := sorted(filenames[0].glob(f"*.{file_type}"), key=text.alpha_numeric_key):
                    self._filenames = list(fns)
                    break
        self._file_type = self._filenames[0].suffix.lstrip(".")

    def get_front_cover(self) -> records.FrontCover | None:
        """Return a FrontCover record or None."""
        for filename in self._filenames:
            one_track = audiofile.AudioFile.open(filename).one_track
            release = one_track.release
            if release.front_cover:
                return release.front_cover
        return None

    def get_search_data(self) -> dict[str, str]:
        """Return a dictionary of search data useful for doing a MusicBrainz search."""
        for filename in self._filenames:
            one_track = audiofile.AudioFile.open(filename).one_track
            release = one_track.release
            track = one_track.track

            artist = track.artist or track.artists.first or "" if track else ""
            album = release.album or "" if release else ""
            mb_artist_id = (
                release.musicbrainz_album_artist_ids.first
                if release and release.musicbrainz_album_artist_ids
                else "" or track.musicbrainz_artist_ids.first
                if track and track.musicbrainz_artist_ids
                else "" or ""
            )
            mb_release_id = release.musicbrainz_album_id or "" if release else ""
            log.info("Artist from tags: %s", artist)
            log.info("Album from tags: %s", album)
            log.info("MB Artist ID from tags: %s", mb_artist_id)
            log.info("MB Release ID from tags: %s", mb_release_id)
            if mb_artist_id and mb_release_id:
                return {"mb_artist_id": mb_artist_id, "mb_release_id": mb_release_id}
            if artist and album:
                return {"artist": artist, "album": album}
        return {}

    def get_source_filenames(self) -> list[pathlib.Path]:
        """Return a list of the original source file paths."""
        return self._filenames

    def prepare_source(self) -> None:
        """Convert the source files to wav files.

        Raises:
             ValueError if the file type is not supported.
        """
        decoders: dict[str, Callable[[str, str], tuple[str, ...]]] = {
            "flac": lambda i, o: ("flac", "--silent", "--decode", f"--output-name={o}", i),
            "m4a": lambda i, o: ("faad", "-q", "-o", o, i),
            "mp3": lambda i, o: ("mpg123", "-q", "-w", o, i),
        }
        try:
            decode = decoders[self._file_type]
        except KeyError as err:
            msg = f"Unsupported source file type: {self._file_type}"
            raise ValueError(msg) from err
        tmp_dir = self._temp_dir / "__tmp__"
        tmp_dir.mkdir(parents=True)
        commands: list[tuple[str, ...]] = []
        for track_number, filepath in enumerate(self.source_list, 1):
            if filepath:
                in_ = str(filepath)
                out_path = tmp_dir / f"{str(track_number).zfill(2)}__.wav"
                out = str(out_path)
                commands.append(decode(in_, out))
                log.info("DECODING: %s -> %s", filepath.name, out_path.name)
        sh.parallel(f"Making {len(commands)} wav files...", commands)
        sh.touch(tmp_dir.glob("*.wav"))
        for filename in sorted(tmp_dir.glob("*.wav"), key=text.alpha_numeric_key):
            subprocess.run(  # noqa: S603
                ("sndfile-convert", "-pcm16", filename, str(filename).replace("/__tmp__/", "/")),
                check=True,
            )
        shutil.rmtree(tmp_dir)



"""Show tags."""

#
#  Copyright (c) 2000-2025 Stephen Jibson
#
#  This file is part of audiolibrarian.
#
#  Audiolibrarian is free software: you can redistribute it and/or modify it under the terms of the
#  GNU General Public License as published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  Audiolibrarian is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
#  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
#  the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with audiolibrarian.
#  If not, see <https://www.gnu.org/licenses/>.
#
import pprint
import sys

import mutagen

if __name__ == "__main__":
    filename = sys.argv[1]

    song = mutagen.File(filename)
    if "covr" in song.tags:
        song.tags["covr"] = ""
    pprint.pp(dict(song.tags))



"""Access the MusicBrainz service."""

#
#  Copyright (c) 2000-2025 Stephen Jibson
#
#  This file is part of audiolibrarian.
#
#  Audiolibrarian is free software: you can redistribute it and/or modify it under the terms of the
#  GNU General Public License as published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  Audiolibrarian is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
#  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
#  the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with audiolibrarian.
#  If not, see <https://www.gnu.org/licenses/>.
#
import dataclasses
import datetime as dt
import http.client
import logging
import pprint
import time
import webbrowser
from typing import TYPE_CHECKING, Any, Final

import musicbrainzngs as mb
import requests
from fuzzywuzzy import fuzz
from requests import auth

from audiolibrarian import __version__, records, text
from audiolibrarian.settings import SETTINGS

if TYPE_CHECKING:
    from collections.abc import Callable

log = logging.getLogger(__name__)
_USER_AGENT_NAME = "audiolibrarian"
_USER_AGENT_CONTACT = "audiolibrarian@jibson.com"
mb.set_useragent(_USER_AGENT_NAME, __version__, _USER_AGENT_CONTACT)


class MusicBrainzSession:
    """MusicBrainzSession provides access to the MusicBrainz API.

    It can be for things that are not supported by the musicbrainzngs library.
    """

    _api_rate = dt.timedelta(seconds=SETTINGS.musicbrainz.rate_limit)
    _last_api_call = dt.datetime.now(tz=dt.UTC)

    def __init__(self) -> None:
        """Initialize a MusicBrainzSession."""
        self.__session: requests.Session | None = None

    def __del__(self) -> None:
        """Close a MusicBrainzSession."""
        if self.__session is not None:
            self.__session.close()

    @property
    def _session(self) -> requests.Session:
        if self.__session is None:
            self.__session = requests.Session()

            if (username := SETTINGS.musicbrainz.username) and (
                password := SETTINGS.musicbrainz.password.get_secret_value()
            ):
                self._session.auth = auth.HTTPDigestAuth(username, password)
            self._session.headers.update(
                {"User-Agent": f"{_USER_AGENT_NAME}/{__version__} ({_USER_AGENT_CONTACT})"}
            )
        return self.__session

    def _get(self, path: str, params: dict[str, str]) -> dict[Any, Any]:
        # Used for direct API calls; those not supported by the python library.
        path = path.lstrip("/")
        url = f"https://musicbrainz.org/ws/2/{path}"
        params["fmt"] = "json"
        self.sleep()
        result = self._session.get(url, params=params)
        while result.status_code == http.HTTPStatus.SERVICE_UNAVAILABLE:
            log.warning("Waiting due to throttling...")
            time.sleep(10)
            result = self._session.get(url, params=params)
        if result.status_code != http.HTTPStatus.OK:
            msg = f"{result.status_code} - {url}"
            raise RuntimeError(msg)
        return dict(result.json())

    def get_artist_by_id(
        self, artist_id: str, includes: list[str] | None = None
    ) -> dict[str, Any]:
        """Return artist for the given musicbrainz-artist ID."""
        params = {}
        if includes is not None:
            params["inc"] = "+".join(includes)
        return self._get(f"artist/{artist_id}", params=params)

    def get_release_group_by_id(
        self, release_group_id: str, includes: list[str] | None = None
    ) -> dict[str, Any]:
        """Return release-group for the given musicbrainz-release-group ID."""
        params = {}
        if includes is not None:
            params["inc"] = "+".join(includes)
        return self._get(f"release-group/{release_group_id}", params=params)

    @staticmethod
    def sleep() -> None:
        """Sleep so we don't abuse the MusicBrainz API service.

        See https://musicbrainz.org/doc/MusicBrainz_API/Rate_Limiting
        """
        since_last = dt.datetime.now(tz=dt.UTC) - MusicBrainzSession._last_api_call
        if (sleep_seconds := (MusicBrainzSession._api_rate - since_last).total_seconds()) > 0:
            log.debug("Sleeping %s to avoid throttling...", sleep_seconds)
            time.sleep(sleep_seconds)
            MusicBrainzSession._last_api_call = dt.datetime.now(tz=dt.UTC)


class MusicBrainzRelease:
    """MusicBrainzRelease reads information from MusicBrains and provides a Release record."""

    _includes: Final[list[str]] = [
        "artist-credits",
        "artist-rels",
        "isrcs",
        "labels",
        "recordings",
        "recording-level-rels",
        "release-groups",
        "tags",
        "work-rels",
        "work-level-rels",
    ]

    def __init__(self, release_id: str, *, verbose: bool = False) -> None:
        """Initialize an MusicBrainzRelease."""
        self._release_id = release_id
        self._verbose = verbose
        self._session = MusicBrainzSession()
        self._session.sleep()
        self._release = mb.get_release_by_id(release_id, includes=self._includes)["release"]
        self._release_record: records.Release | None = None

    def get_release(self) -> records.Release:
        """Return the Release record."""
        if self._release_record is None:
            self._release_record = self._get_release()
        return self._release_record

    def _get_front_cover(self, size: int = 500) -> records.FrontCover | None:
        # Return the FrontCover object (of None).
        if self._release["cover-art-archive"]["front"] == "true":
            self._session.sleep()
            try:
                return records.FrontCover(
                    data=mb.get_image_front(self._release["id"], size=size),
                    desc="front",
                    mime="image/jpeg",
                )
            except (
                mb.musicbrainz.NetworkError,
                mb.musicbrainz.ResponseError,
            ) as err:
                log.warning("Error getting front cover: %s", err)
        return None

    def _get_genre(self, release_group_id: str, artist_id: str) -> str:
        # Try to find the genre using the following methods (in order):
        # - release-group user-genres
        # - artist user-genres
        # - release-group genres
        # - release-group tags
        # - user input

        # user-genres and genres are not supported with the python library.
        release_group = self._session.get_release_group_by_id(
            release_group_id, includes=["genres", "user-genres"]
        )
        log.info("RELEASE_GROUP_GENRES: %s", release_group)
        artist = self._session.get_artist_by_id(artist_id, includes=["genres", "user-genres"])
        log.info("ARTIST_GENRES: %s", artist)
        x_count: Callable[[Any], int] = lambda x: int(x["count"])  # noqa: E731
        if release_group["user-genres"]:
            return str(release_group["user-genres"][0]["name"])
        if artist["user-genres"]:
            return str(artist["user-genres"][0]["name"])
        if release_group["genres"]:
            return str(
                next(g["name"] for g in sorted(release_group["genres"], key=x_count, reverse=True))
            )
        if artist["genres"]:
            return str(
                next(g["name"] for g in sorted(artist["genres"], key=x_count, reverse=True))
            )
        return text.input_("Genre not found; enter the genre [Alternative]: ") or "Alternative"

    def _get_media(self) -> dict[int, records.Medium] | None:
        # Return a dict of Media objects, keyed on number or position (or None).
        media = {}
        for medium in self._release.get("medium-list", []):
            medium_number = int(medium.get("number") or medium.get("position"))
            media[medium_number] = records.Medium(
                formats=records.ListF([medium["format"]]),
                titles=[text.fix(medium["title"])] if medium.get("title") else None,
                track_count=medium["track-count"],
                tracks=self._get_tracks(medium_number=medium_number),
            )
        return media

    def _get_people(self) -> records.People | None:  # noqa: C901, PLR0912
        # Return a People object (or None).
        arrangers, composers, conductors, engineers, lyricists = [], [], [], [], []
        mixers, performers, producers, writers = [], [], [], []
        for relation in self._release.get("artist-relation-list", []):
            if log.getEffectiveLevel() == logging.DEBUG:
                pprint.pp("== ARTIST-RELATION-LIST ===================")
                pprint.pp(relation)
            name = text.fix(relation["artist"]["name"])
            type_ = relation["type"].lower()
            if type_ == "arranger":
                arrangers.append(name)
            elif type_ == "composer":
                composers.append(name)
            elif type_ == "conductor":
                conductors.append(name)
            elif type_ == "engineer":
                engineers.append(name)
            elif type_ == "lyricist":
                lyricists.append(name)
            elif type_ == "mix":
                mixers.append(name)
            elif type_ == "instrument":
                performers.append(
                    records.Performer(
                        name=name, instrument=text.fix(text.join(relation["attribute-list"]))
                    )
                )
            elif type_ == "vocal":
                performers.append(records.Performer(name=name, instrument="lead vocals"))
                if attrs := relation.get("attribute-list"):
                    if attrs := [x for x in attrs if x != "lead vocals"]:
                        performers.append(
                            records.Performer(name=name, instrument=text.fix(text.join(attrs)))
                        )
                else:
                    performers.append(records.Performer(name=name, instrument="vocals"))
            elif type_ == "producer":
                producers.append(name)
            elif type_ == "writer":
                writers.append(name)
            else:
                log.warning("Unknown artist-relation type: %s", type_)
        if (
            engineers
            or arrangers
            or composers
            or conductors
            or lyricists
            or mixers
            or performers
            or producers
            or writers
        ):
            return records.People(
                arrangers=arrangers or None,
                composers=composers or None,
                conductors=conductors or None,
                engineers=engineers or None,
                lyricists=lyricists or None,
                mixers=mixers or None,
                performers=performers or None,
                producers=producers or None,
                writers=writers or None,
            )
        return None

    def _get_release(self) -> records.Release:
        # Return the Release object.
        release = self._release
        log.info("RELEASE %s", release)
        if log.getEffectiveLevel() == logging.DEBUG:
            pprint.pp("== RELEASE ===================")
            pprint.pp(release)
        release_group = release["release-group"]
        (
            album_artist_names_str,
            _,
            album_artist_sort_names,
            artist_ids,
        ) = self._process_artist_credit(release["artist-credit"])
        artist_phrase = text.fix(release.get("artist-credit-phrase", ""))
        year = release.get("release-event-list", [{}])[0].get("date") or text.input_(
            "Release year: "
        )
        album_type = [release_group["primary-type"].lower()]
        if release_group["type"].lower() != album_type[0]:
            album_type.append(release_group["type"].lower())

        labels = list(dict.fromkeys([x["label"]["name"] for x in release["label-info-list"]]))

        key = "catalog-number"
        catalog_numbers = list(
            dict.fromkeys([x[key] for x in release.get("label-info-list", []) if x.get(key)])
        )
        return records.Release(
            album=text.fix(release["title"]),
            album_artists=records.ListF([artist_phrase or album_artist_names_str]),
            album_artists_sort=records.ListF([album_artist_sort_names]),
            asins=[release["asin"]] if release.get("asin") else None,
            barcodes=[release["barcode"]] if release.get("barcode") else None,
            catalog_numbers=catalog_numbers or None,
            date=year,
            front_cover=self._get_front_cover(),
            genres=records.ListF([self._get_genre(release_group["id"], artist_ids.first).title()]),
            labels=labels,
            media=self._get_media(),
            medium_count=release.get("medium-count", 0),
            musicbrainz_album_artist_ids=records.ListF(artist_ids),
            musicbrainz_album_id=self._release_id,
            musicbrainz_release_group_id=release_group["id"],
            original_date=release_group.get("first-release-date", ""),
            original_year=release_group.get("first-release-date", "").split("-")[0] or year,
            people=self._get_people(),
            release_countries=[release.get("country", "")],
            release_statuses=[release.get("status", "").lower()],
            release_types=album_type,
            script="Latn",
            source=records.Source.MUSICBRAINZ,
        )

    def _get_tracks(self, medium_number: int = 1) -> dict[int, records.Track] | None:
        # Return a dict of Track objects, keyed on track number.
        tracks = {}
        for medium in self._release.get("medium-list", []):
            if int(medium["position"]) == medium_number:
                for track in medium["track-list"]:
                    track_number = int(track["position"])
                    recording = track["recording"]
                    artist, artist_list, artist_sort, artist_ids = self._process_artist_credit(
                        track.get("artist-credit") or recording["artist-credit"]
                    )
                    tracks[track_number] = records.Track(
                        artist=artist,
                        artists=artist_list,
                        artists_sort=[artist_sort],
                        isrcs=recording.get("isrc-list"),
                        musicbrainz_artist_ids=artist_ids,
                        musicbrainz_release_track_id=track["id"],
                        musicbrainz_track_id=recording["id"],
                        title=text.fix(track.get("title") or recording["title"]),
                        track_number=track_number,
                    )
                break
        return tracks or None

    @staticmethod
    def _process_artist_credit(
        artist_credit: list[str],
    ) -> tuple[str, records.ListF, str, records.ListF]:
        # Return artist info from an artist-credit list.
        artist_names_str = ""
        artist_names_list = records.ListF()
        artist_sort_names = ""
        artist_ids = records.ListF()
        for credit in artist_credit:
            if isinstance(credit, dict):
                artist_names_str += text.fix(credit.get("name") or credit["artist"]["name"])
                artist_names_list.append(text.fix(credit.get("name") or credit["artist"]["name"]))
                artist_sort_names += text.fix(credit["artist"]["sort-name"])
                artist_ids.append(text.fix(credit["artist"]["id"]))
            else:
                artist_names_str += text.fix(credit)
                artist_sort_names += text.fix(credit)
        return artist_names_str, artist_names_list, artist_sort_names, artist_ids


@dataclasses.dataclass
class Searcher:
    """Searcher objects contain data and methods for searching the MusicBrainz database."""

    artist: str = ""
    album: str = ""
    disc_id: str = ""
    disc_mcn: str = ""
    disc_number: str = ""
    mb_artist_id: str = ""
    mb_release_id: str = ""
    __mb_session: MusicBrainzSession | None = None

    @property
    def _mb_session(self) -> MusicBrainzSession:
        if self.__mb_session is None:
            self.__mb_session = MusicBrainzSession()
        return self.__mb_session

    def find_music_brains_release(self) -> records.Release | None:
        """Return a Release object (or None) based on a search."""
        release_id = self.mb_release_id
        if not release_id and self.disc_id:
            self._mb_session.sleep()
            result = mb.get_releases_by_discid(self.disc_id, includes=["artists"])
            log.info("DISC: {result}")
            if result.get("disc"):
                release_id = result["disc"]["release-list"][0]["id"]
            elif result.get("cdstub"):
                print("A CD Stub exists for this disc, but no disc.")

        if release_id:
            log.info("RELEASE: https://musicbrainz.org/release/%s", release_id)
        elif self.artist and self.album:
            release_group_ids = self._get_release_group_ids()
            log.info("RELEASE_GROUPS: {release_group_ids}")
            release_id = self._prompt_release_id(release_group_ids)
        else:
            release_id = self._prompt_uuid("MusicBrainz Release ID: ")

        return MusicBrainzRelease(release_id).get_release()

    def _get_release_group_ids(self) -> list[str]:
        # Return release groups that fuzzy-match the search criteria.
        artist_l = self.artist.lower()
        album_l = self.album.lower()
        self._mb_session.sleep()
        artist_list = mb.search_artists(query=artist_l, limit=500)["artist-list"]
        if not artist_list:
            return []
        artist_id = artist_list[0]["id"]
        self._mb_session.sleep()
        release_group_list = mb.browse_release_groups(artist=artist_id, limit=500)[
            "release-group-list"
        ]
        log.info("RELEASE_GROUPS: %s", release_group_list)
        if log.getEffectiveLevel() == logging.DEBUG:
            pprint.pp("== RELEASE_GROUPS ===================")
            pprint.pp(release_group_list)
        return [
            rg["id"]
            for rg in release_group_list
            if rg.get("primary-type") == "Album" and fuzz.ratio(album_l, rg["title"].lower()) > 80  # noqa: PLR2004
        ]

    def _prompt_release_id(self, release_group_ids: list[str]) -> str:
        # Prompt for, and return a MusicBrainz release ID.
        print(
            "\n\nWe found the following release group(s). Use the link(s) below to "
            "find the release ID that best matches the audio files.\n"
        )
        for release_group_id in release_group_ids:
            url = f"https://musicbrainz.org/release-group/{release_group_id}"
            print(url)
            webbrowser.open(url)
        return self._prompt_uuid("\nRelease ID or URL: ")

    @staticmethod
    def _prompt_uuid(prompt: str) -> str:
        # Prompt for, and return a UUID.
        while True:
            uuid = text.get_uuid(text.input_(prompt))
            if uuid is not None:
                return uuid



"""Command line commands."""

#
#  Copyright (c) 2000-2025 Stephen Jibson
#
#  This file is part of audiolibrarian.
#
#  Audiolibrarian is free software: you can redistribute it and/or modify it under the terms of the
#  GNU General Public License as published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  Audiolibrarian is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
#  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
#  the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with audiolibrarian.
#  If not, see <https://www.gnu.org/licenses/>.
#
import argparse
import logging
import pathlib
import re
from typing import Any

from audiolibrarian import __version__, audiofile, audiosource, base, genremanager

log = logging.getLogger(__name__)


class _Command:
    # Base class for commands.
    help = ""
    parser = argparse.ArgumentParser()

    @staticmethod
    def validate_args(args: argparse.Namespace) -> bool:
        """Validate command line arguments."""
        _ = args
        return True


class Convert(_Command, base.Base):
    """AudioLibrarian tool for converting and tagging audio files.

    This class performs all of its tasks on instantiation and provides no public members or
    methods.
    """

    command = "convert"
    help = "convert music from files"
    parser = argparse.ArgumentParser()
    parser.add_argument("--artist", "-a", help="provide artist (ignore tags)")
    parser.add_argument("--album", "-m", help="provide album (ignore tags)")
    parser.add_argument("--mb-artist-id", help="MusicBrainz artist ID")
    parser.add_argument("--mb-release-id", help="MusicBrainz release ID")
    parser.add_argument("--disc", "-d", help="format: x/y: disc x of y for multi-disc release")
    parser.add_argument("filename", nargs="+", help="directory name or audio file name")

    def __init__(self, args: argparse.Namespace) -> None:
        """Initialize a Convert command handler."""
        super().__init__(args)
        self._source_is_cd = False
        self._audio_source = audiosource.FilesAudioSource([pathlib.Path(x) for x in args.filename])
        self._get_tag_info()
        self._convert()
        self._write_manifest()

    @staticmethod
    def validate_args(args: argparse.Namespace) -> bool:
        """Validate command line arguments."""
        return _validate_disc_arg(args)


class Genre(_Command):
    """Do stuff with genres."""

    command = "genre"
    help = "manage MB genre"
    parser = argparse.ArgumentParser(
        description=(
            "Process all audio files in the given directory(ies), allowing the user to *update* "
            "the genre in MusicBrainz or *tag* audio files with the user-defined genre in "
            "MusicBrainz."
        )
    )
    parser.add_argument("directory", nargs="+", help="root of directory tree to process")
    parser_action = parser.add_mutually_exclusive_group()
    parser_action.add_argument("--tag", action="store_true", help="update tags")
    parser_action.add_argument("--update", action="store_true", help="update MusicBrainz")

    def __init__(self, args: argparse.Namespace) -> None:
        """Initialize a Genre command handler."""
        genremanager.GenreManager(args)


class Manifest(_Command, base.Base):
    """AudioLibrarian tool for writing manifest.yaml files.

    This class performs all of its tasks on instantiation and provides no public members or
    methods.
    """

    command = "manifest"
    help = "create the Manifest.yaml file"
    parser = argparse.ArgumentParser()
    parser.add_argument("--artist", "-a", help="provide artist (ignore tags)")
    parser.add_argument("--album", "-m", help="provide album (ignore tags)")
    parser.add_argument("--cd", "-c", action="store_true", help="original source was a CD")
    parser.add_argument("--mb-artist-id", help="MusicBrainz artist ID")
    parser.add_argument("--mb-release-id", help="MusicBrainz release ID")
    parser.add_argument("--disc", "-d", help="format: x/y: disc x of y for multi-disc release")
    parser.add_argument("filename", nargs="+", help="directory name or audio file name")

    def __init__(self, args: argparse.Namespace) -> None:
        """Initialize a Manifest command handler."""
        super().__init__(args)
        self._source_is_cd = args.cd
        self._audio_source = audiosource.FilesAudioSource([pathlib.Path(x) for x in args.filename])
        source_filenames = self._audio_source.get_source_filenames()
        self._source_example = audiofile.AudioFile.open(source_filenames[0]).read_tags()
        self._get_tag_info()
        self._write_manifest()

    @staticmethod
    def validate_args(args: argparse.Namespace) -> bool:
        """Validate command line arguments."""
        return _validate_disc_arg(args)


class Reconvert(_Command, base.Base):
    """AudioLibrarian tool for re-converting and tagging audio files from existing source files.

    This class performs all of its tasks on instantiation and provides no public members or
    methods.
    """

    command = "reconvert"
    help = "re-convert files from an existing source directory"
    parser = argparse.ArgumentParser()
    parser.add_argument("directories", nargs="+", help="source directories")

    def __init__(self, args: argparse.Namespace) -> None:
        """Initialize a Reconvert command handler."""
        super().__init__(args)
        self._source_is_cd = False
        manifest_paths = self._find_manifests(args.directories)
        count = len(manifest_paths)
        for i, manifest_path in enumerate(manifest_paths):
            print(f"Processing {i + 1} of {count} ({i / count:.0%}): {manifest_path}...")
            self._audio_source = audiosource.FilesAudioSource([manifest_path.parent])
            manifest = self._read_manifest(manifest_path)
            self._disc_number, self._disc_count = manifest["disc_number"], manifest["disc_count"]
            self._get_tag_info()
            self._convert(make_source=False)

    @staticmethod
    def validate_args(args: argparse.Namespace) -> bool:
        """Validate command line arguments."""
        return _validate_directories_arg(args)


class Rename(_Command, base.Base):
    """AudioLibrarian tool for renaming audio files based on their tags.

    This class performs all of its tasks on instantiation and provides no public members or
    methods.
    """

    command = "rename"
    help = "rename files based on tags or MusicBrainz data"
    parser = argparse.ArgumentParser()
    parser.add_argument("--dry-run", action="store_true", help="don't actually rename files")
    parser.add_argument("directories", nargs="+", help="audio file directories")

    def __init__(self, args: argparse.Namespace) -> None:
        """Initialize a Rename command handler."""
        super().__init__(args)
        self._source_is_cd = False
        print("Finding audio files...")
        for audio_file in self._find_audio_files(args.directories):
            filepath = audio_file.filepath
            if audio_file.one_track.track is None:
                log.warning("%s has no title", filepath)
                continue
            depth = 3 if filepath.parent.name.startswith("disc") else 2
            old_name = filepath
            new_name = (
                filepath.parents[depth]
                / audio_file.one_track.get_artist_album_disc_path()
                / audio_file.one_track.track.get_filename(filepath.suffix)
            )
            if old_name != new_name:
                print(f"Renaming:\n  {old_name} -> \n  {new_name}")
                if args.dry_run:
                    continue
                old_parent = old_name.parent
                new_parent = new_name.parent
                new_parent.mkdir(parents=True, exist_ok=True)
                old_name.rename(new_name)
                if not old_parent.samefile(new_parent):
                    # Move the Manifest if it's the only file left.
                    man = "Manifest.yaml"
                    if [f.name for f in old_parent.glob("*")] == [man]:
                        print(f"Renaming:\n  {old_parent / man} -> \n  {new_parent / man}")
                        (old_parent / man).rename(new_parent / man)
                    for idx in range(depth):
                        if not list(old_name.parents[idx].glob("*")):
                            print(f"Removing: {old_name.parents[idx]}")
                            old_name.parents[idx].rmdir()

            else:
                log.debug("Not renaming %s", filepath)

    @staticmethod
    def validate_args(args: argparse.Namespace) -> bool:
        """Validate command line arguments."""
        return _validate_directories_arg(args)


class Rip(_Command, base.Base):
    """AudioLibrarian tool for ripping, converting and tagging audio files.

    This class performs all of its tasks on instantiation and provides no public members or
    methods.
    """

    command = "rip"
    help = "rip music from a CD"
    parser = argparse.ArgumentParser()
    parser.add_argument("--artist", "-a", help="provide artist")
    parser.add_argument("--album", "-m", help="provide album")
    parser.add_argument("--mb-artist-id", help="MusicBrainz artist ID")
    parser.add_argument("--mb-release-id", help="MusicBrainz release ID")
    parser.add_argument("--disc", "-d", help="x/y: disc x of y; multi-disc release")

    def __init__(self, args: argparse.Namespace) -> None:
        """Initialize a Rip command handler."""
        super().__init__(args)
        self._source_is_cd = True
        self._audio_source = audiosource.CDAudioSource()
        self._get_tag_info()
        self._convert()
        self._write_manifest()

    @staticmethod
    def validate_args(args: argparse.Namespace) -> bool:
        """Validate command line arguments."""
        return _validate_disc_arg(args)


class Version(_Command):
    """Print the version."""

    command = "version"
    help = "display the program version"

    def __init__(self, args: argparse.Namespace) -> None:
        """Initialize a Version command handler."""
        _ = args
        print(f"audiolibrarian {__version__}")


def _validate_directories_arg(args: argparse.Namespace) -> bool:
    for directory in args.directories:
        if not pathlib.Path(directory).is_dir():
            print(f"Directory not found: {directory}")
            return False
    return True


def _validate_disc_arg(args: argparse.Namespace) -> bool:
    if "disc" in args and args.disc:
        if not re.match(r"\d+/\d+", args.disc):
            print("Invalid --disc specification; should be 'x/y'")
            return False
        x, y = args.disc.split("/")
        if int(x) > int(y) or int(x) < 1 or int(y) < 1:
            print("Invalid --disc specification; should be 'x/y' where x <= y and x and y  >= 1")
            return False
    return True


COMMANDS: set[Any] = {Convert, Genre, Manifest, Reconvert, Rename, Rip, Version}



"""Record (dataclass) definitions.

Useful field reference: https://github.com/metabrainz/picard/blob/master/picard/util/tags.py
"""

#
#  Copyright (c) 2000-2025 Stephen Jibson
#
#  This file is part of audiolibrarian.
#
#  Audiolibrarian is free software: you can redistribute it and/or modify it under the terms of the
#  GNU General Public License as published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  Audiolibrarian is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
#  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
#  the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with audiolibrarian.
#  If not, see <https://www.gnu.org/licenses/>.
#
import dataclasses
import enum
import pathlib
from typing import Any

from audiolibrarian import text


class BitrateMode(enum.Enum):
    """Bitrate modes."""

    UNKNOWN = 0
    CBR = 1
    VBR = 2


class FileType(enum.Enum):
    """Audio file types."""

    UNKNOWN = 0
    AAC = 1
    FLAC = 2
    MP3 = 3
    WAV = 4


class Source(enum.Enum):
    """Information source."""

    MUSICBRAINZ = 1
    TAGS = 2


class ListF(list[Any]):
    """A list, with a first property."""

    @property
    def first(self) -> Any | None:  # noqa: ANN401
        """Return the first element in the list (or None, if the list is empty)."""
        return self[0] if self else None


@dataclasses.dataclass(kw_only=True)
class Record:
    """Base class for records.

    Overrides true/false test for records returning false if all fields are None.
    """

    def __bool__(self) -> bool:
        """Return a boolean representation of the Record."""
        return bool([x for x in dataclasses.asdict(self).values() if x is not None])

    def asdict(self) -> dict[Any, Any]:
        """Return a dict version of the record."""
        return dataclasses.asdict(self)


# Primitive Record Types
@dataclasses.dataclass(kw_only=True)
class FileInfo(Record):
    """File information."""

    bitrate: int | None = None
    bitrate_mode: BitrateMode | None = None
    path: pathlib.Path | None = None
    type: FileType | None = None


@dataclasses.dataclass(kw_only=True)
class FrontCover(Record):
    """A front cover."""

    data: bytes | None = None
    desc: str | None = None
    mime: str | None = None


@dataclasses.dataclass(kw_only=True)
class Performer(Record):
    """A performer with an instrument."""

    name: str | None = None
    instrument: str | None = None


@dataclasses.dataclass(kw_only=True)
class Track(Record):
    """A track."""

    artist: str | None = None
    artists: ListF | None = None
    artists_sort: list[str] | None = None
    file_info: FileInfo | None = None
    isrcs: list[str] | None = None
    musicbrainz_artist_ids: ListF | None = None
    musicbrainz_release_track_id: str | None = None
    musicbrainz_track_id: str | None = None
    title: str | None = None
    track_number: int | None = None

    def get_filename(self, suffix: str = "") -> str:
        """Return a sane filename based on track number and title.

        If suffix is included, it will be appended to the filename.
        """
        if self.title is None or self.track_number is None:
            msg = "Unable to generate a filename for Track with missing number and/or title"
            raise ValueError(msg)
        return (
            str(self.track_number).zfill(2) + "__" + text.filename_from_title(self.title) + suffix
        )


# Combined Record Types (fields + other record types)
@dataclasses.dataclass(kw_only=True)
class Medium(Record):
    """A medium."""

    formats: ListF | None = None
    titles: list[str] | None = None
    track_count: int | None = None
    tracks: dict[int, Track] | None = None


@dataclasses.dataclass(kw_only=True)
class People(Record):
    """People."""

    arrangers: list[str] | None = None
    composers: list[str] | None = None
    conductors: list[str] | None = None
    engineers: list[str] | None = None
    lyricists: list[str] | None = None
    mixers: list[str] | None = None
    performers: list[Performer] | None = None
    producers: list[str] | None = None
    writers: list[str] | None = None


@dataclasses.dataclass(kw_only=True)
class Release(Record):
    """A release."""

    album: str | None = None
    album_artists: ListF | None = None
    album_artists_sort: ListF | None = None
    asins: list[str] | None = None
    barcodes: list[str] | None = None
    catalog_numbers: list[str] | None = None
    date: str | None = None
    front_cover: FrontCover | None = dataclasses.field(default=None, repr=False)
    genres: ListF | None = None
    labels: list[str] | None = None
    media: dict[int, Medium] | None = None
    medium_count: int | None = None
    musicbrainz_album_artist_ids: ListF | None = None
    musicbrainz_album_id: str | None = None
    musicbrainz_release_group_id: str | None = None
    original_date: str | None = None
    original_year: str | None = None
    people: People | None = None
    release_countries: list[str] | None = None
    release_statuses: list[str] | None = None
    release_types: list[str] | None = None
    script: str | None = None
    source: Source | None = None

    def get_artist_album_path(self) -> pathlib.Path:
        """Return a directory for the artist/album/disc combination.

        Example:
          -  artist__the/1969__the_album
        """
        if self.album_artists_sort is None or self.album_artists_sort.first is None:
            msg = "Unable to determine artist path without artist(s)"
            raise ValueError(msg)
        artist_dir = pathlib.Path(text.filename_from_title(self.album_artists_sort.first))
        if self.original_year is None or self.album is None:
            msg = "Unable to determine album path without year and album"
            raise ValueError(msg)
        album_dir = pathlib.Path(text.filename_from_title(f"{self.original_year}__{self.album}"))
        return artist_dir / album_dir

    def pp(self, medium_number: int) -> str:
        """Return a string summary of the Release."""
        if self.media is None:
            msg = "Missing release information"
            raise ValueError(msg)
        tracks = "\n".join(
            (
                f"  {str(n).zfill(2)}: {t.title}"
                for n, t in sorted(self.media[medium_number].tracks.items())
            )
        )
        return "\n".join(
            (
                f"Album: {self.album}",
                f"Artist(s): {', '.join(self.album_artists)}",
                f"Medium: {medium_number} of {self.medium_count}",
                "Tracks:",
                tracks,
            )
        )


@dataclasses.dataclass(kw_only=True)
class OneTrack(Record):
    """A single track."""

    release: Release | None = None
    medium_number: int | None = None
    track_number: int | None = None

    @property
    def medium(self) -> Medium | None:
        """Return the Medium object (or None)."""
        if self.release and self.release.media:
            return self.release.media[self.medium_number]
        return None

    @property
    def track(self) -> Track | None:
        """Return the Track object (or None)."""
        if self.medium and self.medium.tracks:
            return self.medium.tracks[self.track_number]
        return None

    def get_artist_album_disc_path(self) -> pathlib.Path:
        """Return a directory for the artist/album/disc combination.

        Example:
          - artist__the/1969__the_album
          - artist__the/1969__the_album/disc2
        """
        if (self.medium_number, self.release.medium_count) == (1, 1):
            return self.release.get_artist_album_path()
        return self.release.get_artist_album_path() / f"disc{self.medium_number}"



"""Genre Manager."""

#
#  Copyright (c) 2000-2025 Stephen Jibson
#
#  This file is part of audiolibrarian.
#
#  Audiolibrarian is free software: you can redistribute it and/or modify it under the terms of the
#  GNU General Public License as published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  Audiolibrarian is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
#  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
#  the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with audiolibrarian.
#  If not, see <https://www.gnu.org/licenses/>.
#
import argparse
import logging
import pathlib
import pickle
import typing
import webbrowser
from typing import Any

import mutagen
import mutagen.flac
import mutagen.id3
import mutagen.mp4

from audiolibrarian import musicbrainz, text
from audiolibrarian.settings import SETTINGS

log = logging.getLogger(__name__)


class GenreManager:
    """Manage genres."""

    def __init__(self, args: argparse.Namespace) -> None:
        """Initialize a GenreManager instance."""
        self._args = args
        self._mb = musicbrainz.MusicBrainzSession()
        self._paths = self._get_all_paths()
        self._paths_by_artist = self._get_paths_by_artist()
        _u, _c = self._get_genres_by_artist()
        self._user_genres_by_artist, self._community_genres_by_artist = _u, _c
        if self._args.update:
            self._update_user_artists()
        elif self._args.tag:
            self._update_tags()  # type: ignore[no-untyped-call]

    @typing.no_type_check  # The mutagen library doesn't provide type hints.
    def _update_tags(self) -> None:
        """Set the genre tags for all songs to the user-based genre."""
        for artist_id, paths in self._paths_by_artist.items():
            genre = self._user_genres_by_artist.get(artist_id)
            if not genre:
                continue
            for path in paths:
                if path.suffix == ".flac":
                    flac_song = mutagen.flac.FLAC(str(path))
                    current_genre = flac_song.tags["genre"][0]
                    if current_genre != genre:
                        flac_song.tags["genre"] = genre
                        flac_song.save()
                        print(f"{path}: {current_genre} --> {genre}")
                elif path.suffix == ".m4a":
                    m4a_song = mutagen.mp4.MP4(str(path))
                    current_genre = m4a_song.tags["\xa9gen"][0]
                    if current_genre != genre:
                        m4a_song.tags["\xa9gen"] = genre
                        m4a_song.save()
                        print(f"{path}: {current_genre} --> {genre}")
                elif path.suffix == ".mp3":
                    mp3_song = mutagen.File(str(path))
                    current_genre = str(mp3_song.tags["TCON"])
                    if current_genre != genre:
                        id3 = mutagen.id3.ID3(str(path))
                        id3.add(mutagen.id3.TCON(encoding=3, text=genre))
                        id3.save()
                        print(f"{path}: {current_genre} --> {genre}")

    def _update_user_artists(self) -> None:
        """Pull up a web page to allow the user to set the genre.

        Update the genre for all artists that only have community-base genres.
        """
        for artist_id, artist in sorted(
            self._community_genres_by_artist.items(),
            key=lambda x: x[1]["name"],
        ):
            artist_name = artist["name"]
            i = (
                text.input_(f"Continue with {artist_name} (YES, no, skip)[Y, n, s]: ")
                .lower()
                .strip()
            )
            if i == "n":
                break
            if i != "s":
                webbrowser.open(f"https://musicbrainz.org/artist/{artist_id}/tags")

    def _get_all_paths(self) -> list[pathlib.Path]:
        """Return a list of paths for all files in the directories specified in the args."""
        paths = []
        for directory in self._args.directory:
            paths.extend([p for p in list(pathlib.Path(directory).glob("**/*")) if p.is_file()])
        return paths

    def _get_paths_by_artist(self) -> dict[str, list[pathlib.Path]]:
        """Return a map of artist-IDs to paths representing audio files by that artist."""
        artists: dict[str, list[pathlib.Path]] = {}
        for path in self._paths:
            artist_id = None
            song = mutagen.File(str(path))
            if path.suffix == ".flac":
                artist_id = str(
                    song.tags.get("musicbrainz_albumartistid", [""])[0]
                    or song.tags.get("musicbrainz_artistid", [""])[0]
                )
            elif path.suffix == ".m4a":
                artist_id = (
                    song.tags.get("----:com.apple.iTunes:MusicBrainz Album Artist Id", [""])[0]
                    or song.tags.get("----:com.apple.iTunes:MusicBrainz Artist Id", [""])[0]
                ).decode("utf8")
            elif path.suffix == ".mp3":
                artist_id = str(
                    song.tags.get("TXXX:MusicBrainz Album Artist Id", [""])[0]
                    or song.tags.get("TXXX:MusicBrainz Artist Id", [""])[0]
                )
            if artist_id:
                if artist_id not in artists:
                    artists[artist_id] = []
                artists[artist_id].append(path)
        return artists

    def _get_genres_by_artist(
        self,
    ) -> tuple[dict[str, str], dict[str, dict[str, Any]]]:
        """Return two dicts mapping MusicBrainz-artist-ID to user and community.

        Returns:
          user: a single genre, set in MusicBrainz by this app's user
          community: a list genre records (dicts) set in MusicBrainz by the community
                     with "name" and "count" fields
        """
        user: dict[str, str] = {}
        community: dict[str, dict[str, Any]] = {}
        user_modified = False
        cache_file = SETTINGS.work_dir / "user-genres.pickle"
        if cache_file.exists():
            with cache_file.open(mode="rb") as cache_file_obj:
                user = pickle.load(cache_file_obj)  # noqa: S301
        for artist_id in self._paths_by_artist:
            if artist_id in user:
                log.debug("Cache hit: %s %s", artist_id, user[artist_id])
                continue  # Already in the cache.
            artist = self._mb.get_artist_by_id(artist_id, includes=["genres", "user-genres"])
            if artist["user-genres"]:
                genre = artist["user-genres"][0]["name"].title()
                user[artist_id] = genre
                user_modified = True
            elif artist["genres"]:
                community[artist_id] = {
                    "name": artist["name"],
                    "genres": [{"name": x["name"], "count": x["count"]} for x in artist["genres"]],
                }
        if user_modified:
            cache_file.parent.mkdir(exist_ok=True)
            with cache_file.open(mode="wb") as cache_file_obj:
                # noinspection PyTypeChecker
                pickle.dump(user, cache_file_obj)

        return user, community



"""Configuration module for AudioLibrarian.

This module provides configuration management using pydantic-settings, supporting multiple
sources of configuration:

1. Environment Variables (highest precedence):
   - Prefix: "AUDIOLIBRARIAN__"
   - Nested fields: Use "__" as delimiter (e.g., AUDIOLIBRARIAN__MUSICBRAINZ__USERNAME)

2. YAML Configuration File:
   - Location: $XDG_CONFIG_HOME/audiolibrarian/config.yaml
   - Supports nested structure for MusicBrainz settings

3. Default Values:
   - Defined in Settings class
   - Lowest precedence

The configuration is immutable (frozen=True) and follows XDG base directory standards for
configuration and cache locations.

Sensitive fields (like passwords) are handled using pydantic.SecretStr for security.
"""

#
#  Copyright (c) 2000-2025 Stephen Jibson
#
#  This file is part of audiolibrarian.
#
#  Audiolibrarian is free software: you can redistribute it and/or modify it under the terms of the
#  GNU General Public License as published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  Audiolibrarian is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
#  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
#  the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with audiolibrarian.
#  If not, see <https://www.gnu.org/licenses/>.
#
import logging
import pathlib
from typing import Literal

import pydantic
import pydantic_settings
import xdg_base_dirs

logger = logging.getLogger(__name__)


class MusicBrainzSettings(pydantic_settings.BaseSettings):
    """Configuration settings for MusicBrainz."""

    password: pydantic.SecretStr = pydantic.SecretStr("")
    username: str = ""
    rate_limit: pydantic.PositiveFloat = 1.5  # Seconds between requests.


class NormalizeFFmpegSettings(pydantic_settings.BaseSettings):
    """Configuration settings for ffmpeg normalization."""

    target_level: float = -13


class NormalizeWavegainSettings(pydantic_settings.BaseSettings):
    """Configuration settings for wavegain normalization."""

    gain: int = 5  # dB
    preset: Literal["album", "radio"] = "radio"


class NormalizeSettings(pydantic_settings.BaseSettings):
    """Configuration settings for audio normalization."""

    normalizer: Literal["auto", "wavegain", "ffmpeg", "none"] = "auto"
    ffmpeg: NormalizeFFmpegSettings = NormalizeFFmpegSettings()
    wavegain: NormalizeWavegainSettings = NormalizeWavegainSettings()


class Settings(pydantic_settings.BaseSettings):
    """Configuration settings for AudioLibrarian."""

    discid_device: str | None = None  # Use default device.
    library_dir: pathlib.Path = pathlib.Path("library").resolve()
    musicbrainz: MusicBrainzSettings = MusicBrainzSettings()
    normalize: NormalizeSettings = NormalizeSettings()
    work_dir: pathlib.Path = xdg_base_dirs.xdg_cache_home() / "audiolibrarian"

    model_config = pydantic_settings.SettingsConfigDict(
        env_nested_delimiter="__",
        env_prefix="AUDIOLIBRARIAN__",
        yaml_file=str(xdg_base_dirs.xdg_config_home() / "audiolibrarian" / "config.yaml"),
        frozen=True,  # Make settings immutable.
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[pydantic_settings.BaseSettings],
        init_settings: pydantic_settings.PydanticBaseSettingsSource,
        env_settings: pydantic_settings.PydanticBaseSettingsSource,
        dotenv_settings: pydantic_settings.PydanticBaseSettingsSource,
        file_secret_settings: pydantic_settings.PydanticBaseSettingsSource,
    ) -> tuple[pydantic_settings.PydanticBaseSettingsSource, ...]:
        del dotenv_settings, file_secret_settings  # Unused.
        return (
            env_settings,
            pydantic_settings.YamlConfigSettingsSource(settings_cls),
            init_settings,
        )


SETTINGS = Settings()
__all__ = ["SETTINGS"]



"""Audio file library."""

#
#  Copyright (c) 2000-2025 Stephen Jibson
#
#  This file is part of audiolibrarian.
#
#  Audiolibrarian is free software: you can redistribute it and/or modify it under the terms of the
#  GNU General Public License as published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  Audiolibrarian is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
#  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
#  the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with audiolibrarian.
#  If not, see <https://www.gnu.org/licenses/>.
#
import abc
import importlib
import pathlib
from typing import Any, ClassVar

import mutagen

from audiolibrarian import records


class AudioFile(abc.ABC):
    """Abstract base class for AudioFile classes."""

    _subclass_by_extension: ClassVar[dict[str, type["AudioFile"]]] = {}

    def __init__(self, filepath: pathlib.Path) -> None:
        """Initialize an AudioFile."""
        self._filepath = filepath
        self._mut_file = mutagen.File(self.filepath.absolute())
        self._one_track = self.read_tags()

    def __init_subclass__(cls, extensions: set[str], **kwargs: dict[str, Any]) -> None:
        """Initialize an AudioFile subclass."""
        super().__init_subclass__(**kwargs)
        for extension in extensions:
            cls._subclass_by_extension[extension] = cls

    def __repr__(self) -> str:
        """Return a string representation of the AudioFile."""
        return f"AudioFile: {self.filepath}"

    @classmethod
    def extensions(cls) -> set[str]:
        """Return the list of supported extensions."""
        return set(cls._subclass_by_extension.keys())

    @classmethod
    def open(cls, filename: str | pathlib.Path) -> "AudioFile":
        """Return an AudioFile object based on the filename extension (factory method).

        Args:
            filename: The filename of a supported audio file.

        Returns:
            audiofile.AudioFile: An AudioFile object.

        Raises:
            FileNotFoundError: If the file cannot be found or is not a file.
            NotImplementedError: If the type of the file is not supported.
        """
        if not AudioFile._subclass_by_extension:
            # Dynamically load submodules.
            for module_path in (pathlib.Path(__file__).parent / "formats").glob("*.py"):
                if module_path.name == "__init__.py":
                    continue
                importlib.import_module(f"audiolibrarian.audiofile.formats.{module_path.stem}")

        filepath = pathlib.Path(filename).resolve()
        if not filepath.is_file():
            raise FileNotFoundError(filepath)
        if filepath.suffix not in AudioFile._subclass_by_extension:
            msg = f"Unknown file type: {filepath}"
            raise NotImplementedError(msg)
        return AudioFile._subclass_by_extension[filepath.suffix](filepath=filepath)

    @property
    def filepath(self) -> pathlib.Path:
        """Return the audio file's path."""
        return self._filepath

    @property
    def one_track(self) -> records.OneTrack:
        """Return the OneTrack representation of the audio file."""
        return self._one_track

    @one_track.setter
    def one_track(self, one_track: records.OneTrack) -> None:
        """Set the OneTrack representation of the audio file."""
        self._one_track = one_track

    @abc.abstractmethod
    def read_tags(self) -> records.OneTrack:
        """Read the tags from the audio file and return a populated OneTrack record."""

    @abc.abstractmethod
    def write_tags(self) -> None:
        """Write the tags to the audio file."""

    def _get_tag_sources(self) -> tuple[records.Release, int, records.Medium, int, records.Track]:
        # Return the objects and information required to generate tags.
        release = self.one_track.release or records.Release()
        medium_number = self.one_track.medium_number
        medium = self.one_track.medium or records.Medium()
        track_number = self.one_track.track_number
        track = self.one_track.track or records.Track()
        return release, medium_number, medium, track_number, track



"""Text utilities."""

#
#  Copyright (c) 2000-2025 Stephen Jibson
#
#  This file is part of audiolibrarian.
#
#  Audiolibrarian is free software: you can redistribute it and/or modify it under the terms of the
#  GNU General Public License as published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  Audiolibrarian is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
#  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
#  the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with audiolibrarian.
#  If not, see <https://www.gnu.org/licenses/>.
#
import pathlib
import re
import sys
from typing import Any

import picard_src

_DIGIT_REGEX = re.compile(r"([0-9]+)")
_UUID_REGEX = re.compile(
    r"[a-f0-9]{8}-?[a-f0-9]{4}-?[a-f0-9]{4}-?[a-f0-9]{4}-?[a-f0-9]{12}", re.IGNORECASE
)


def alpha_numeric_key(text: str | pathlib.Path) -> list[Any]:
    """Return a key that can be used for sorting alphanumeric strings numerically.

    Example:
        from audiolibrarian import text

        l = ["8__eight", "7__seven", "10__ten", "11__eleven"]
        sorted(l)
        # ['10__ten', '11__eleven', '7__seven', '8__eight']
        sorted(l, key=text.alpha_numeric_key)
        # ['7__seven', '8__eight', '10__ten', '11__eleven']
    """
    return [int(x) if x.isdigit() else x for x in _DIGIT_REGEX.split(str(text))]


def filename_from_title(title: str) -> str:
    """Convert a title into a filename."""
    escape_required = "'!\"#$&'()*;<>?[]\\`{}|~\t\n); "
    invalid = "/"
    no_underscore_replace = "'!\""
    results: list[str] = []
    for char in picard_src.replace_non_ascii(title):  # type: ignore[no-untyped-call]
        if char == "&":
            results.extend("and")
        elif char.isascii() and char not in escape_required and char not in invalid:
            results.append(char)
        elif char not in no_underscore_replace:
            results.append("_")
    result = "".join(results).rstrip("_")
    # Strip tailing dots, unless we end with an upper-case letter, then put one dot back.
    if result.endswith(".") and result.rstrip(".")[-1].isupper():
        return result.rstrip(".") + "."
    return result.rstrip(".")


def fix(text: str) -> str:
    """Replace some special characters."""
    text = picard_src.unicode_simplify_combinations(text)  # type: ignore[no-untyped-call]
    text = picard_src.unicode_simplify_punctuation(text)  # type: ignore[no-untyped-call]
    return picard_src.unicode_simplify_compatibility(text)  # type: ignore[no-untyped-call, no-any-return]


def get_numbers(text: str) -> list[int]:
    """Get a list of all the numbers in the given string."""
    return [int(x) for x in _DIGIT_REGEX.findall(text)]


def get_track_number(filename: str) -> int:
    """Get a track number from a filename or from the user."""
    if numbers := get_numbers(filename):
        return numbers[0]
    while True:  # pragma: no cover
        try:
            return int(input_(f"Enter the track number for: {filename}"))
        except ValueError:
            pass


def get_uuid(text: str) -> str | None:
    """Return the first UUID found within a given string."""
    if (match := _UUID_REGEX.search(text)) is not None:
        return match.group()
    return None


def input_(prompt: str) -> str:  # pragma: no cover
    """Sound a terminal bell then prompt the user for input."""
    sys.stdout.write("\a")  # Terminal bell escape char.
    sys.stdout.flush()
    return input(prompt)


def join(strings: list[str], joiner: str = ", ", word: str = "and") -> str:
    """Join string with joiner and word.

    Example:
        text.join(["eggs", "bacon", "spam"])
        # "eggs, bacon and spam"
    """
    if not strings:
        return ""
    if len(strings) == 1:
        return strings[0]
    return joiner.join(strings[:-1]) + " " + word + " " + strings[-1]



"""Audiolibrarian command line interface."""

#
#  Copyright (c) 2000-2025 Stephen Jibson
#
#  This file is part of audiolibrarian.
#
#  Audiolibrarian is free software: you can redistribute it and/or modify it under the terms of the
#  GNU General Public License as published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  Audiolibrarian is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
#  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
#  the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with audiolibrarian.
#  If not, see <https://www.gnu.org/licenses/>.
#
import argparse
import logging
import pathlib
import subprocess
import sys
from typing import Final

from audiolibrarian import commands

log = logging.getLogger("audiolibrarian")


class CommandLineInterface:
    """Command line interface."""

    _REQUIRED_EXE: Final[set[str]] = {
        "cd-paranoia",
        "eject",
        "faad",
        "fdkaac",
        "flac",
        "lame",
        "mpg123",
        "sndfile-convert",
    }

    def __init__(self, *, parse_args: bool = True) -> None:
        """Initialize a CommandLineInterface handler."""
        if parse_args:
            self._args = self._parse_args()
            self.log_level = self._args.log_level

    def execute(self) -> None:
        """Execute the command."""
        log.info("ARGS: %s", self._args)
        if not self._check_deps():
            sys.exit(1)
        for cmd in commands.COMMANDS:
            if self._args.command == cmd.command:
                if not cmd.validate_args(self._args):
                    sys.exit(2)
                cmd(self._args)
                break
        if self._args.log_level == logging.DEBUG:
            print(pathlib.Path("/proc/self/status").read_text(encoding="utf-8"))

    def _check_deps(self) -> bool:
        """Check that all the executables defined in REQUIRED_EXE exist on the system.

        If any of the required executables are missing, list them and return False.
        """
        missing = []
        for exe in self._REQUIRED_EXE:
            try:
                subprocess.run(  # noqa: S602
                    f"command -v {exe}",
                    shell=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                    check=True,
                )
            except subprocess.CalledProcessError:
                missing.append(exe)
        if missing:
            print(f"\nMissing required executable(s): {', '.join(missing)}\n")
            return False
        return True

    # noinspection PyProtectedMember
    @staticmethod
    def _parse_args() -> argparse.Namespace:
        """Parse command line arguments."""
        parser = argparse.ArgumentParser(description="audiolibrarian")

        # global options
        parser.add_argument(
            "--log-level",
            "-l",
            choices=("CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"),
            default="ERROR",
            help="log level (default: ERROR)",
        )

        # Add sub-commands and args for sub_commands.
        subparsers = parser.add_subparsers(title="commands", dest="command")
        for cmd_ in commands.COMMANDS:
            # This is a total hack because argparse won't allow you to add an already
            # existing ArgumentParser as a sub-parser.
            if cmd_.parser:
                cmd_.parser.prog = f"{subparsers._prog_prefix} {cmd_.command}"  # noqa: SLF001
                subparsers._choices_actions.append(  # noqa: SLF001
                    subparsers._ChoicesPseudoAction(cmd_.command, (), cmd_.help)  # noqa: SLF001
                )
                subparsers._name_parser_map[cmd_.command] = cmd_.parser  # noqa: SLF001

        return parser.parse_args()


def main() -> None:
    """Execute the command line interface."""
    cli_ = CommandLineInterface()
    logging.basicConfig(level=cli_.log_level)
    logging.captureWarnings(capture=True)
    cli_.execute()



"""AudioFile formats."""
#
#  Copyright (c) 2000-2025 Stephen Jibson
#
#  This file is part of audiolibrarian.
#
#  Audiolibrarian is free software: you can redistribute it and/or modify it under the terms of the
#  GNU General Public License as published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  Audiolibrarian is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
#  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
#  the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with audiolibrarian.
#  If not, see <https://www.gnu.org/licenses/>.
#
